from typing import Optional
from typing import Tuple

import torch
from torch import Tensor

from pfhedge._utils.typing import TensorOrScalar

from ._utils import cast_state


def _get_epsilon(dtype: Optional[torch.dtype]) -> float:
    return torch.finfo(dtype).tiny if dtype else torch.finfo().tiny


def generate_cir(
    n_paths: int,
    n_steps: int,
    init_state: Optional[Tuple[TensorOrScalar, ...]] = None,
    kappa: TensorOrScalar = 1.0,
    theta: TensorOrScalar = 0.04,
    sigma: TensorOrScalar = 0.2,
    dt: TensorOrScalar = 1 / 250,
    dtype: Optional[torch.dtype] = None,
    device: Optional[torch.device] = None,
) -> Tensor:
    r"""Returns time series following Cox-Ingersoll-Ross process.

    The time evolution of the process is given by:

    .. math::

        dX(t) = \kappa (\theta - X(t)) dt + \sigma \sqrt{X(t)} dW(t) .

    Time series is generated by Andersen's QE-M method (See Reference for details).

    References:
        - Andersen, Leif B.G., Efficient Simulation of the Heston Stochastic
          Volatility Model (January 23, 2007). Available at SSRN:
          https://ssrn.com/abstract=946405 or http://dx.doi.org/10.2139/ssrn.946404

    Args:
        n_paths (int): The number of simulated paths.
        n_steps (int): The number of time steps.
        init_state (tuple[torch.Tensor | float], optional): The initial state of
            the time series.
            This is specified by a tuple :math:`(X(0),)`.
            It also accepts a :class:`torch.Tensor` or a :class:`float`.
            If ``None`` (default), it uses :math:`(\theta, )`.
        kappa (torch.Tensor or float, default=1.0): The parameter :math:`\kappa`.
        theta (torch.Tensor or float, default=0.04): The parameter :math:`\theta`.
        sigma (torch.Tensor or float, default=2.0): The parameter :math:`\sigma`.
        dt (torch.Tensor or float, default=1/250): The intervals of the time steps.
        dtype (torch.dtype, optional): The desired data type of returned tensor.
            Default: If ``None``, uses a global default
            (see :func:`torch.set_default_tensor_type()`).
        device (torch.device, optional): The desired device of returned tensor.
            Default: if None, uses the current device for the default tensor type
            (see :func:`torch.set_default_tensor_type()`).
            ``device`` will be the CPU for CPU tensor types and the current CUDA device
            for CUDA tensor types.

    Shape:
        - Output: :math:`(N, T)` where
          :math:`N` is the number of paths and
          :math:`T` is the number of time steps.

    Returns:
        torch.Tensor

    Examples:
        >>> from pfhedge.stochastic import generate_cir
        >>>
        >>> _ = torch.manual_seed(42)
        >>> generate_cir(2, 5)
        tensor([[0.0400, 0.0408, 0.0411, 0.0417, 0.0422],
                [0.0400, 0.0395, 0.0452, 0.0434, 0.0446]])
    """
    if init_state is None:
        init_state = (theta,)

    init_state = cast_state(init_state, dtype=dtype, device=device)

    # PSI_CRIT in [1.0, 2.0]. See section 3.2.3
    PSI_CRIT = 1.5
    # Prevent zero division
    EPSILON = _get_epsilon(dtype)

    output = torch.empty(*(n_paths, n_steps), dtype=dtype, device=device)  # type: ignore
    output[:, 0] = init_state[0]

    randn = torch.randn_like(output)
    rand = torch.rand_like(output)

    # Cast to Tensor with desired dtype and device
    kappa, theta, sigma, dt = map(torch.as_tensor, (kappa, theta, sigma, dt))
    kappa, theta, sigma, dt = map(lambda t: t.to(output), (kappa, theta, sigma, dt))

    for i_step in range(n_steps - 1):
        v = output[:, i_step]

        # Compute m, s, psi: Eq(17,18)
        exp = (-kappa * dt).exp()
        m = theta + (v - theta) * exp
        s2 = v * (sigma ** 2) * exp * (1 - exp) / kappa + theta * (sigma ** 2) * (
            (1 - exp).square()
        ) / (2 * kappa)
        psi = s2 / m.square().clamp(min=EPSILON)

        # Compute V(t + dt) where psi <= PSI_CRIT: Eq(23, 27, 28)
        b = ((2 / psi) - 1 + (2 / psi).sqrt() * (2 / psi - 1).sqrt()).sqrt()
        a = m / (1 + b.square())
        next_0 = a * (b + randn[:, i_step]).square()

        # Compute V(t + dt) where psi > PSI_CRIT: Eq(25)
        u = rand[:, i_step]
        p = (psi - 1) / (psi + 1)
        beta = (1 - p) / m.clamp(min=EPSILON)
        pinv = ((1 - p) / (1 - u).clamp(min=EPSILON)).log() / beta
        next_1 = torch.where(u > p, pinv, torch.zeros_like(u))

        output[:, i_step + 1] = torch.where(psi <= PSI_CRIT, next_0, next_1)

    return output



from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import List
from typing import Optional
from typing import TypeVar
from typing import no_type_check

import torch
from torch import Tensor

T = TypeVar("T", bound="BaseInstrument")


class BaseInstrument(ABC):
    """Base class for all financial instruments."""

    cost: float

    @property
    @abstractmethod
    def spot(self) -> Tensor:
        """Returns the spot price of self."""

    @property
    @abstractmethod
    def is_listed(self) -> bool:
        """Returns whether it is listed or not."""

    @abstractmethod
    @no_type_check
    def simulate(self, n_paths: int, time_horizon: float, **kwargs) -> None:
        """Simulate time series associated with the instrument itself
        (for a primary instrument) or its underlier (for a derivative)
        and add them as buffers.

        Args:
            n_paths (int): The number of paths to simulate.
            time_horizon (float): The period of time to simulate the price.

        Returns:
            None
        """

    @abstractmethod
    def to(self: T, *args: Any, **kwargs: Any) -> T:
        """Moves and/or casts the buffers of the instrument.

        This can be called as

        .. function:: to(device=None, dtype=None)
        .. function:: to(tensor)
        .. function:: to(instrument)

        Its signature is similar to :meth:`torch.nn.Module.to`.
        It only accepts floating point dtypes.
        See :ref:`instrument-attributes-doc` for details.

        Note:
            This method modifies the instrument in-place.

        .. seealso::
            - :meth:`float()`: Cast to :class:`torch.float32`.
            - :meth:`double()`: Cast to :class:`torch.float64`.
            - :meth:`half()`: Cast to :class:`torch.float16`.
            - :meth:`bfloat16()`: Cast to :class:`torch.bfloat16`.
            - :meth:`cuda()`: Move to CUDA memory.
            - :meth:`cpu()`: Move to CPU memory.

        Args:
            dtype (torch.dtype): The desired floating point dtype of
                the buffers in this instrument.
            device (torch.device): The desired device of
                the buffers in this instrument.
            tensor (torch.Tensor): Tensor whose dtype and device are
                the desired dtype and device of
                the buffers in this instrument.
            instrument (BaseInstrument): Instrument whose dtype and device are
                the desired dtype and device of
                the buffers in this instrument.

        Returns:
            self
        """

    def cpu(self: T) -> T:
        """Moves all buffers of this instrument and its underlier to the CPU.

        Note:
            This method modifies the instrument in-place.

        Returns:
            self
        """
        return self.to(torch.device("cpu"))

    def cuda(self: T, device: Optional[int] = None) -> T:
        """Moves all buffers of this instrument and its underlier to the GPU.

        Note:
            This method modifies the instrument in-place.

        Args:
            device (int, optional): If specified,
                all buffers will be copied to that device.

        Returns:
            self
        """
        return self.to(torch.device(f"cuda:{device}" if device is not None else "cuda"))

    def double(self: T) -> T:
        """Casts all floating point parameters and buffers to
        ``torch.float64`` datatype.

        Note:
            This method modifies the instrument in-place.

        Returns:
            self
        """
        return self.to(torch.float64)

    def float(self: T) -> T:
        """Casts all floating point parameters and buffers to
        ``torch.float32`` datatype.

        Note:
            This method modifies the instrument in-place.

        Returns:
            self
        """
        return self.to(torch.float32)

    def half(self: T) -> T:
        """Casts all floating point parameters and buffers to
        ``torch.float16`` datatype.

        Note:
            This method modifies the instrument in-place.

        Returns:
            self
        """
        return self.to(torch.float16)

    def float64(self: T) -> T:
        """Alias for :meth:`double()`."""
        return self.double()

    def float32(self: T) -> T:
        """Alias for :meth:`float()`."""
        return self.float()

    def float16(self: T) -> T:
        """Alias for :meth:`half()`."""
        return self.half()

    def bfloat16(self: T) -> T:
        """Casts all floating point parameters and buffers to
        ``torch.bfloat16`` datatype.

        Note:
            This method modifies the instrument in-place.

        Returns:
            self
        """
        return self.to(torch.bfloat16)

    def extra_repr(self) -> str:
        """Set the extra representation of the instrument.

        To print customized extra information,
        you should re-implement this method in your own instruments.
        Both single-line and multi-line strings are acceptable.
        """
        return ""

    def _get_name(self) -> str:
        return self.__class__.__name__

    def _dinfo(self) -> List[str]:
        # Returns list of strings that tell ``dtype`` and ``device`` of self.
        # Intended to be used in :func:`__repr__`.
        # If ``dtype`` (``device``) is the one specified in default type,
        # ``dinfo`` will not have the information of it.
        # Implementation here refers to the function _str_intern in
        # pytorch/_tensor_str.py.
        dinfo = []

        dtype = getattr(self, "dtype", None)
        if dtype is not None:
            if dtype != torch.get_default_dtype():
                dinfo.append("dtype=" + str(dtype))

        # A general logic here is we only print device when it doesn't match
        # the device specified in default tensor type.
        device = getattr(self, "device", None)
        if device is not None:
            if device.type != torch._C._get_default_device() or (
                device.type == "cuda" and torch.cuda.current_device() != device.index
            ):
                dinfo.append("device='" + str(device) + "'")

        return dinfo


class Instrument(BaseInstrument):
    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)  # type: ignore
        raise DeprecationWarning(
            "Instrument is deprecated. Use BaseInstrument instead."
        )



from typing import Optional

import torch
from torch import Tensor

from pfhedge._utils.bisect import find_implied_volatility
from pfhedge._utils.doc import _set_attr_and_docstring
from pfhedge._utils.str import _format_float
from pfhedge.instruments import AmericanBinaryOption
from pfhedge.nn.functional import bs_american_binary_delta
from pfhedge.nn.functional import bs_american_binary_price

from ._base import BSModuleMixin
from ._base import acquire_params_from_derivative_0
from ._base import acquire_params_from_derivative_2
from .black_scholes import BlackScholesModuleFactory


class BSAmericanBinaryOption(BSModuleMixin):
    """Black-Scholes formula for an American binary option.

    Note:
        - The formulas are for continuous monitoring while
          :class:`pfhedge.instruments.AmericanBinaryOption` monitors spot prices discretely.
          To get adjustment for discrete monitoring, see, for instance,
          Broadie, Glasserman, and Kou (1999).

    Args:
        call (bool, default=True): Specifies whether the option is call or put.
        strike (float, default=1.0): The strike price of the option.

    Shape:
        - Input: :math:`(N, *, 4)` where
          :math:`*` means any number of additional dimensions.
          See :meth:`inputs` for the names of input features.
        - Output: :math:`(N, *, 1)`.
          All but the last dimension are the same shape as the input.

    .. seealso::
        - :class:`pfhedge.nn.BlackScholes`:
          Initialize Black-Scholes formula module from a derivative.
        - :class:`pfhedge.instruments.AmericanBinaryOption`:
          Corresponding derivative.

    References:
        - Shreve, S.E., 2004. Stochastic calculus for finance II:
          Continuous-time models (Vol. 11). Springer Science & Business Media.
        - Broadie, M., Glasserman, P. and Kou, S.G., 1999.
          Connecting discrete and continuous path-dependent options.
          Finance and Stochastics, 3(1), pp.55-82.

    Examples:
        >>> from pfhedge.nn import BSAmericanBinaryOption
        >>>
        >>> m = BSAmericanBinaryOption(strike=1.0)
        >>> m.inputs()
        ['log_moneyness', 'max_log_moneyness', 'time_to_maturity', 'volatility']
        >>> input = torch.tensor([
        ...     [-0.01, -0.01, 0.1, 0.2],
        ...     [ 0.00,  0.00, 0.1, 0.2],
        ...     [ 0.01,  0.01, 0.1, 0.2]])
        >>> m(input)
        tensor([[...],
                [...],
                [...]])
    """

    def __init__(
        self,
        call: bool = True,
        strike: float = 1.0,
        derivative: Optional[AmericanBinaryOption] = None,
    ):
        if not call:
            raise ValueError(
                f"{self.__class__.__name__} for a put option is not yet supported."
            )

        super().__init__()
        self.call = call
        self.strike = strike
        self.derivative = derivative

    @classmethod
    def from_derivative(cls, derivative):
        """Initialize a module from a derivative.

        Args:
            derivative (:class:`pfhedge.instruments.AmericanBinaryOption`):
                The derivative to get the Black-Scholes formula.

        Returns:
            BSAmericanBinaryOption

        Examples:
            >>> from pfhedge.instruments import BrownianStock
            >>> from pfhedge.instruments import AmericanBinaryOption
            >>>
            >>> derivative = AmericanBinaryOption(BrownianStock(), strike=1.1)
            >>> m = BSAmericanBinaryOption.from_derivative(derivative)
            >>> m
            BSAmericanBinaryOption(strike=1.1000)
        """
        return cls(
            call=derivative.call, strike=derivative.strike, derivative=derivative
        )

    def extra_repr(self) -> str:
        params = []
        params.append("strike=" + _format_float(self.strike))
        return ", ".join(params)

    def price(
        self,
        log_moneyness: Optional[Tensor] = None,
        max_log_moneyness: Optional[Tensor] = None,
        time_to_maturity: Optional[Tensor] = None,
        volatility: Optional[Tensor] = None,
    ) -> Tensor:
        """Returns price of the derivative.

        Args:
            log_moneyness (torch.Tensor, optional): Log moneyness of the underlying asset.
            max_log_moneyness (torch.Tensor, optional): Cumulative maximum of the log moneyness.
            time_to_maturity (torch.Tensor, optional): Time to expiry of the option.
            volatility (torch.Tensor, optional): Volatility of the underlying asset.

        Shape:
            - log_moneyness: :math:`(N, *)` where
              :math:`*` means any number of additional dimensions.
            - max_log_moneyness: :math:`(N, *)`
            - time_to_maturity: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            torch.Tensor

        Note:
            Parameters are not optional if the module has not accepted a derivative in its initialization.
        """
        # This formula is derived using the results in Section 7.3.3 of Shreve's book.
        # Price is I_2 - I_4 where the interval of integration is [k --> -inf, b].
        # By this substitution we get N([log(S(0) / K) + ...] / sigma T) --> 1.
        (
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        ) = acquire_params_from_derivative_2(
            self.derivative,
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        )
        return bs_american_binary_price(
            log_moneyness=log_moneyness,
            max_log_moneyness=max_log_moneyness,
            time_to_maturity=time_to_maturity,
            volatility=volatility,
        )

    def delta(
        self,
        log_moneyness: Optional[Tensor] = None,
        max_log_moneyness: Optional[Tensor] = None,
        time_to_maturity: Optional[Tensor] = None,
        volatility: Optional[Tensor] = None,
    ) -> Tensor:
        """Returns delta of the derivative.

        Args:
            log_moneyness (torch.Tensor, optional): Log moneyness of the underlying asset.
            max_log_moneyness (torch.Tensor, optional): Cumulative maximum of the log moneyness.
            time_to_maturity (torch.Tensor, optional): Time to expiry of the option.
            volatility (torch.Tensor, optional): Volatility of the underlying asset.

        Shape:
            - log_moneyness: :math:`(N, *)` where
              :math:`*` means any number of additional dimensions.
            - max_log_moneyness: :math:`(N, *)`
            - time_to_maturity: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            torch.Tensor

        Note:
            Parameters are not optional if the module has not accepted a derivative in its initialization.
        """
        (
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        ) = acquire_params_from_derivative_2(
            self.derivative,
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        )
        return bs_american_binary_delta(
            log_moneyness=log_moneyness,
            max_log_moneyness=max_log_moneyness,
            time_to_maturity=time_to_maturity,
            volatility=volatility,
            strike=self.strike,
        )

    @torch.enable_grad()
    def gamma(
        self,
        log_moneyness: Optional[Tensor] = None,
        max_log_moneyness: Optional[Tensor] = None,
        time_to_maturity: Optional[Tensor] = None,
        volatility: Optional[Tensor] = None,
    ) -> Tensor:
        """Returns gamma of the derivative.

        Args:
            log_moneyness (torch.Tensor, optional): Log moneyness of the underlying asset.
            max_log_moneyness (torch.Tensor, optional): Cumulative maximum of the log moneyness.
            time_to_maturity (torch.Tensor, optional): Time to expiry of the option.
            volatility (torch.Tensor, optional): Volatility of the underlying asset.

        Shape:
            - log_moneyness: :math:`(N, *)` where
              :math:`*` means any number of additional dimensions.
            - max_log_moneyness: :math:`(N, *)`
            - time_to_maturity: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            torch.Tensor

        Note:
            args are not optional if it doesn't accept derivative in this initialization.
        """
        (
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        ) = acquire_params_from_derivative_2(
            self.derivative,
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        )
        return super().gamma(
            log_moneyness=log_moneyness,
            max_log_moneyness=max_log_moneyness,
            time_to_maturity=time_to_maturity,
            volatility=volatility,
            strike=self.strike,
        )

    @torch.enable_grad()
    def vega(
        self,
        log_moneyness: Optional[Tensor] = None,
        max_log_moneyness: Optional[Tensor] = None,
        time_to_maturity: Optional[Tensor] = None,
        volatility: Optional[Tensor] = None,
    ) -> Tensor:
        """Returns vega of the derivative.

        Args:
            log_moneyness (torch.Tensor, optional): Log moneyness of the underlying asset.
            max_log_moneyness (torch.Tensor, optional): Cumulative maximum of the log moneyness.
            time_to_maturity (torch.Tensor, optional): Time to expiry of the option.
            volatility (torch.Tensor, optional): Volatility of the underlying asset.

        Shape:
            - log_moneyness: :math:`(N, *)` where
              :math:`*` means any number of additional dimensions.
            - max_log_moneyness: :math:`(N, *)`
            - time_to_maturity: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            torch.Tensor

        Note:
            args are not optional if it doesn't accept derivative in this initialization.
        """
        (
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        ) = acquire_params_from_derivative_2(
            self.derivative,
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        )
        return super().vega(
            log_moneyness=log_moneyness,
            max_log_moneyness=max_log_moneyness,
            time_to_maturity=time_to_maturity,
            volatility=volatility,
            strike=self.strike,
        )

    @torch.enable_grad()
    def theta(
        self,
        log_moneyness: Optional[Tensor] = None,
        max_log_moneyness: Optional[Tensor] = None,
        time_to_maturity: Optional[Tensor] = None,
        volatility: Optional[Tensor] = None,
    ) -> Tensor:
        """Returns theta of the derivative.

        Args:
            log_moneyness (torch.Tensor, optional): Log moneyness of the underlying asset.
            max_log_moneyness (torch.Tensor, optional): Cumulative maximum of the log moneyness.
            time_to_maturity (torch.Tensor, optional): Time to expiry of the option.
            volatility (torch.Tensor, optional): Volatility of the underlying asset.

        Shape:
            - log_moneyness: :math:`(N, *)` where
              :math:`*` means any number of additional dimensions.
            - max_log_moneyness: :math:`(N, *)`
            - time_to_maturity: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Note:
            Risk-free rate is set to zero.

        Returns:
            torch.Tensor

        Note:
            args are not optional if it doesn't accept derivative in this initialization.
        """
        (
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        ) = acquire_params_from_derivative_2(
            self.derivative,
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        )
        return super().theta(
            log_moneyness=log_moneyness,
            max_log_moneyness=max_log_moneyness,
            time_to_maturity=time_to_maturity,
            volatility=volatility,
            strike=self.strike,
        )

    def implied_volatility(
        self,
        log_moneyness: Optional[Tensor] = None,
        max_log_moneyness: Optional[Tensor] = None,
        time_to_maturity: Optional[Tensor] = None,
        price: Optional[Tensor] = None,
        precision: float = 1e-6,
    ) -> Tensor:
        """Returns implied volatility of the derivative.

        Args:
            log_moneyness (torch.Tensor, optional): Log moneyness of the underlying asset.
            max_log_moneyness (torch.Tensor, optional): Cumulative maximum of the log moneyness.
            time_to_maturity (torch.Tensor, optional): Time to expiry of the option.
            volatility (torch.Tensor, optional): Volatility of the underlying asset.
            price (torch.Tensor): Price of the derivative.
            precision (float, default=1e-6): Computational precision of the implied volatility.

        Shape:
            - log_moneyness: :math:`(N, *)` where
              :math:`*` means any number of additional dimensions.
            - max_log_moneyness: :math:`(N, *)`
            - time_to_maturity: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            torch.Tensor

        Note:
            args are not optional if it doesn't accept derivative in this initialization.
            price seems optional in typing, but it isn't. It is set for the compatibility to the previous versions.
        """
        (log_moneyness, time_to_maturity) = acquire_params_from_derivative_0(
            self.derivative, log_moneyness, time_to_maturity
        )
        if price is None:
            raise ValueError(
                "price is required in this method. None is set only for compatibility to the previous versions."
            )
        return find_implied_volatility(
            self.price,
            price=price,
            log_moneyness=log_moneyness,
            max_log_moneyness=max_log_moneyness,
            time_to_maturity=time_to_maturity,
            precision=precision,
        )


factory = BlackScholesModuleFactory()
factory.register_module("AmericanBinaryOption", BSAmericanBinaryOption)

# Assign docstrings so they appear in Sphinx documentation
_set_attr_and_docstring(BSAmericanBinaryOption, "inputs", BSModuleMixin.inputs)
_set_attr_and_docstring(BSAmericanBinaryOption, "forward", BSModuleMixin.forward)



from typing import List
from typing import Optional

import torch
from torch import Tensor

from pfhedge._utils.bisect import find_implied_volatility
from pfhedge._utils.doc import _set_attr_and_docstring
from pfhedge._utils.doc import _set_docstring
from pfhedge._utils.str import _format_float
from pfhedge.instruments import LookbackOption
from pfhedge.nn.functional import bs_lookback_price

from ._base import BSModuleMixin
from ._base import acquire_params_from_derivative_0
from ._base import acquire_params_from_derivative_2
from .black_scholes import BlackScholesModuleFactory


class BSLookbackOption(BSModuleMixin):
    """Black-Scholes formula for a lookback option with a fixed strike.

    Note:
        - The formulas are for continuous monitoring while
          :class:`pfhedge.instruments.LookbackOption` monitors spot prices discretely.
          To get adjustment for discrete monitoring, see, for instance,
          Broadie, Glasserman, and Kou (1999).

    .. seealso::
        - :class:`pfhedge.nn.BlackScholes`:
          Initialize Black-Scholes formula module from a derivative.
        - :class:`pfhedge.instruments.LookbackOption`:
          Corresponding derivative.

    References:
        - Conze, A., 1991. Path dependent options: The case of lookback options.
          The Journal of Finance, 46(5), pp.1893-1907.
        - Broadie, M., Glasserman, P. and Kou, S.G., 1999.
          Connecting discrete and continuous path-dependent options.
          Finance and Stochastics, 3(1), pp.55-82.

    Args:
        call (bool, default=True): Specifies whether the option is call or put.
        strike (float, default=1.0): The strike price of the option.

    Shape:
        - Input: :math:`(N, *, 4)` where
          :math:`*` means any number of additional dimensions.
          See :meth:`inputs` for the names of input features.
        - Output: :math:`(N, *, 1)`.
          All but the last dimension are the same shape as the input.

    Examples:
        >>> from pfhedge.nn import BSLookbackOption
        >>>
        >>> m = BSLookbackOption()
        >>> m.inputs()
        ['log_moneyness', 'max_log_moneyness', 'time_to_maturity', 'volatility']
        >>> input = torch.tensor([
        ...     [-0.01, -0.01, 0.1, 0.2],
        ...     [ 0.00,  0.00, 0.1, 0.2],
        ...     [ 0.01,  0.01, 0.1, 0.2]])
        >>> m(input)
        tensor([[0.9208],
                [1.0515],
                [1.0515]])
    """

    def __init__(
        self,
        call: bool = True,
        strike: float = 1.0,
        derivative: Optional[LookbackOption] = None,
    ) -> None:
        if not call:
            raise ValueError(
                f"{self.__class__.__name__} for a put option is not yet supported."
            )

        super().__init__()
        self.call = call
        self.strike = strike
        self.derivative = derivative

    @classmethod
    def from_derivative(cls, derivative):
        """Initialize a module from a derivative.

        Args:
            derivative (:class:`pfhedge.instruments.LookbackOption`):
                The derivative to get the Black-Scholes formula.

        Returns:
            BSLookbackOption

        Examples:
            >>> from pfhedge.instruments import BrownianStock
            >>> from pfhedge.instruments import LookbackOption
            >>>
            >>> derivative = LookbackOption(BrownianStock(), strike=1.1)
            >>> m = BSLookbackOption.from_derivative(derivative)
            >>> m
            BSLookbackOption(strike=1.1000)
        """
        return cls(
            call=derivative.call, strike=derivative.strike, derivative=derivative
        )

    def extra_repr(self) -> str:
        params = []
        params.append("strike=" + _format_float(self.strike))
        return ", ".join(params)

    def inputs(self) -> List[str]:
        return ["log_moneyness", "max_log_moneyness", "time_to_maturity", "volatility"]

    def price(
        self,
        log_moneyness: Optional[Tensor] = None,
        max_log_moneyness: Optional[Tensor] = None,
        time_to_maturity: Optional[Tensor] = None,
        volatility: Optional[Tensor] = None,
    ) -> Tensor:
        r"""Returns price of the derivative.

        The price is given by:

        .. math::
            \begin{cases}
                S(0) \{
                    N(d_1) + \sigma \sqrt{T} [N'(d_1) + d_1 N(d_1)]
                \} - K N(d_2)
                & (M \leq K) \\
                S(0) \{
                    N(d_1') + \sigma \sqrt{T} [N'(d_1') + d_1' N(d_1')]
                \} - K + M [1 - N(d_2')]
                & (M > K) \\
            \end{cases}
        where
        :math:`M = \max_{t < 0} S(t)`,
        :math:`d_1' = [\log(S(0) / M) + \frac12 \sigma^2 T] / \sigma \sqrt{T}`, and
        :math:`d_2' = [\log(S(0) / M) - \frac12 \sigma^2 T] / \sigma \sqrt{T}`.

        Args:
            log_moneyness (torch.Tensor, optional): Log moneyness of the underlying asset.
            max_log_moneyness (torch.Tensor, optional): Cumulative maximum of the log moneyness.
            time_to_maturity (torch.Tensor, optional): Time to expiry of the option.
            volatility (torch.Tensor, optional): Volatility of the underlying asset.

        Shape:
            - log_moneyness: :math:`(N, *)` where
              :math:`*` means any number of additional dimensions.
            - max_log_moneyness: :math:`(N, *)`
            - time_to_maturity: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            torch.Tensor

        Note:
            Parameters are not optional if the module has not accepted a derivative in its initialization.
        """
        (
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        ) = acquire_params_from_derivative_2(
            self.derivative,
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        )
        return bs_lookback_price(
            log_moneyness=log_moneyness,
            max_log_moneyness=max_log_moneyness,
            time_to_maturity=time_to_maturity,
            volatility=volatility,
            strike=self.strike,
        )

    @torch.enable_grad()
    def delta(
        self,
        log_moneyness: Optional[Tensor] = None,
        max_log_moneyness: Optional[Tensor] = None,
        time_to_maturity: Optional[Tensor] = None,
        volatility: Optional[Tensor] = None,
        create_graph: bool = False,
    ) -> Tensor:
        """Returns delta of the derivative.

        Args:
            log_moneyness (torch.Tensor, optional): Log moneyness of the underlying asset.
            max_log_moneyness (torch.Tensor, optional): Cumulative maximum of the log moneyness.
            time_to_maturity (torch.Tensor, optional): Time to expiry of the option.
            volatility (torch.Tensor, optional): Volatility of the underlying asset.
            create_graph (bool, default=False): If True, graph of the derivative
                will be constructed. This option is used to compute gamma.

        Shape:
            - log_moneyness: :math:`(N, *)` where
              :math:`*` means any number of additional dimensions.
            - max_log_moneyness: :math:`(N, *)`
            - time_to_maturity: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            torch.Tensor

        Note:
            Parameters are not optional if the module has not accepted a derivative in its initialization.
        """
        (
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        ) = acquire_params_from_derivative_2(
            self.derivative,
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        )
        return super().delta(
            log_moneyness=log_moneyness,
            max_log_moneyness=max_log_moneyness,
            time_to_maturity=time_to_maturity,
            volatility=volatility,
            create_graph=create_graph,
            strike=self.strike,
        )

    @torch.enable_grad()
    def gamma(
        self,
        log_moneyness: Optional[Tensor] = None,
        max_log_moneyness: Optional[Tensor] = None,
        time_to_maturity: Optional[Tensor] = None,
        volatility: Optional[Tensor] = None,
    ) -> Tensor:
        """Returns gamma of the derivative.

        Args:
            log_moneyness (torch.Tensor, optional): Log moneyness of the underlying asset.
            max_log_moneyness (torch.Tensor, optional): Cumulative maximum of the log moneyness.
            time_to_maturity (torch.Tensor, optional): Time to expiry of the option.
            volatility (torch.Tensor, optional):
                Volatility of the underlying asset.

        Shape:
            - log_moneyness: :math:`(N, *)` where
              :math:`*` means any number of additional dimensions.
            - max_log_moneyness: :math:`(N, *)`
            - time_to_maturity: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            torch.Tensor

        Note:
            Arguments are not optional if it doesn't accept derivative in this initialization.
        """
        (
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        ) = acquire_params_from_derivative_2(
            self.derivative,
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        )
        return super().gamma(
            strike=self.strike,
            log_moneyness=log_moneyness,
            max_log_moneyness=max_log_moneyness,
            time_to_maturity=time_to_maturity,
            volatility=volatility,
        )

    @torch.enable_grad()
    def vega(
        self,
        log_moneyness: Optional[Tensor] = None,
        max_log_moneyness: Optional[Tensor] = None,
        time_to_maturity: Optional[Tensor] = None,
        volatility: Optional[Tensor] = None,
    ) -> Tensor:
        """Returns vega of the derivative.

        Args:
            log_moneyness (torch.Tensor, optional): Log moneyness of the underlying asset.
            max_log_moneyness (torch.Tensor, optional): Cumulative maximum of the log moneyness.
            time_to_maturity (torch.Tensor, optional): Time to expiry of the option.
            volatility (torch.Tensor, optional):
                Volatility of the underlying asset.

        Shape:
            - log_moneyness: :math:`(N, *)` where
              :math:`*` means any number of additional dimensions.
            - max_log_moneyness: :math:`(N, *)`
            - time_to_maturity: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            torch.Tensor

        Note:
            - Arguments are not optional if it doesn't accept derivative in this initialization.
        """
        (
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        ) = acquire_params_from_derivative_2(
            self.derivative,
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        )
        return super().vega(
            strike=self.strike,
            log_moneyness=log_moneyness,
            max_log_moneyness=max_log_moneyness,
            time_to_maturity=time_to_maturity,
            volatility=volatility,
        )

    @torch.enable_grad()
    def theta(
        self,
        log_moneyness: Optional[Tensor] = None,
        max_log_moneyness: Optional[Tensor] = None,
        time_to_maturity: Optional[Tensor] = None,
        volatility: Optional[Tensor] = None,
    ) -> Tensor:
        """Returns theta of the derivative.

        Args:
            log_moneyness (torch.Tensor, optional): Log moneyness of the underlying asset.
            max_log_moneyness (torch.Tensor, optional): Cumulative maximum of the log moneyness.
            time_to_maturity (torch.Tensor, optional): Time to expiry of the option.
            volatility (torch.Tensor, optional):
                Volatility of the underlying asset.

        Shape:
            - log_moneyness: :math:`(N, *)` where
              :math:`*` means any number of additional dimensions.
            - max_log_moneyness: :math:`(N, *)`
            - time_to_maturity: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            torch.Tensor

        Note:
            - Risk-free rate is set to zero.
            - Arguments are not optional if it doesn't accept derivative in this initialization.
        """
        (
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        ) = acquire_params_from_derivative_2(
            self.derivative,
            log_moneyness,
            max_log_moneyness,
            time_to_maturity,
            volatility,
        )
        return super().theta(
            strike=self.strike,
            log_moneyness=log_moneyness,
            max_log_moneyness=max_log_moneyness,
            time_to_maturity=time_to_maturity,
            volatility=volatility,
        )

    def implied_volatility(
        self,
        log_moneyness: Optional[Tensor] = None,
        max_log_moneyness: Optional[Tensor] = None,
        time_to_maturity: Optional[Tensor] = None,
        price: Optional[Tensor] = None,
        precision: float = 1e-6,
    ) -> Tensor:
        """Returns implied volatility of the derivative.

        Args:
            log_moneyness (torch.Tensor, optional): Log moneyness of the underlying asset.
            max_log_moneyness (torch.Tensor, optional): Cumulative maximum of the log moneyness.
            time_to_maturity (torch.Tensor, optional): Time to expiry of the option.
            price (torch.Tensor): Price of the derivative.
            precision (float, default=1e-6): Precision of the implied volatility.

        Shape:
            - log_moneyness: :math:`(N, *)` where
              :math:`*` means any number of additional dimensions.
            - max_log_moneyness: :math:`(N, *)`
            - time_to_maturity: :math:`(N, *)`
            - price: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            torch.Tensor

        Note:
            - Arguments are not optional if it doesn't accept derivative in this initialization.
        """
        # price seems optional in typing, but it isn't. It is set for the compatibility to the previous versions.
        (log_moneyness, time_to_maturity) = acquire_params_from_derivative_0(
            self.derivative, log_moneyness, time_to_maturity
        )
        if price is None:
            raise ValueError(
                "price is required in this method. None is set only for compatibility to the previous versions."
            )
        return find_implied_volatility(
            self.price,
            price=price,
            log_moneyness=log_moneyness,
            max_log_moneyness=max_log_moneyness,
            time_to_maturity=time_to_maturity,
            precision=precision,
        )


factory = BlackScholesModuleFactory()
factory.register_module("LookbackOption", BSLookbackOption)

# Assign docstrings so they appear in Sphinx documentation
_set_docstring(BSLookbackOption, "inputs", BSModuleMixin.inputs)
_set_attr_and_docstring(BSLookbackOption, "forward", BSModuleMixin.forward)



from typing import Optional
from typing import Tuple

import torch
from torch import Tensor
from torch.quasirandom import SobolEngine

from pfhedge.nn.functional import box_muller


def _get_numel(size: Tuple[int, ...]) -> int:
    out = 1
    for dim in size:
        out *= dim
    return out


class RandnSobolBoxMuller:
    """Generator of random numbers from a standard normal distribution
    using Sobol sequence and Box-Muller transformation.

    Args:
        scramble (bool, optional): Setting this to ``True`` will produce
            scrambled Sobol sequences. Scrambling is capable of producing
            better Sobol sequences. Default: ``False``.
        seed (int, optional): This is the seed for the scrambling.
            The seed of the random number generator is set to this,
            if specified. Otherwise, it uses a random seed.
            Default: ``None``.
    """

    def __init__(self, scramble: bool = False, seed: Optional[int] = None):
        self.scramble = scramble
        self.seed = seed

    def __call__(
        self,
        *size: int,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
    ) -> Tensor:
        numel = _get_numel(tuple(size))
        output = self._generate_1d(numel, dtype=dtype, device=device)
        output.resize_(*size)
        return output

    def _generate_1d(
        self,
        n: int,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
    ) -> Tensor:
        engine = SobolEngine(2, scramble=self.scramble, seed=self.seed)
        rand = engine.draw(n // 2 + 1).to(dtype=dtype, device=device)
        z0, z1 = box_muller(rand[:, 0], rand[:, 1])
        return torch.cat((z0, z1), dim=0)[:n]



from inspect import signature
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from typing import no_type_check

from torch import Tensor
from torch.nn import Module

import pfhedge.autogreek as autogreek
from pfhedge.instruments import AmericanBinaryOption
from pfhedge.instruments import EuropeanBinaryOption
from pfhedge.instruments import EuropeanOption
from pfhedge.instruments import LookbackOption


class BSModuleMixin(Module):
    """A mixin class for Black-Scholes formula modules.

    Shape:
        - Input: :math:`(N, *, H_\\text{in})` where
          :math:`H_\\text{in}` is the number of input features and
          :math:`*` means any number of additional dimensions.
          See :meth:`inputs` for the names of input features.
        - Output: :math:`(N, *, 1)`.
          All but the last dimension are the same shape as the input.
    """

    def forward(self, input: Tensor) -> Tensor:
        """Returns delta of the derivative.

        Args:
            input (torch.Tensor): The input tensor. Features are concatenated along
                the last dimension.
                See :meth:`inputs()` for the names of the input features.

        Returns:
            torch.Tensor
        """
        return self.delta(*(input[..., [i]] for i in range(input.size(-1))))

    @no_type_check
    def price(self, *args, **kwargs) -> Tensor:
        """Returns price of the derivative.

        Returns:
            torch.Tensor
        """

    @no_type_check
    def delta(self, **kwargs) -> Tensor:
        """Returns delta of the derivative.

        Returns:
            torch.Tensor
        """
        return autogreek.delta(self.price, **kwargs)

    @no_type_check
    def gamma(self, **kwargs) -> Tensor:
        """Returns delta of the derivative.

        Returns:
            torch.Tensor
        """
        return autogreek.gamma(self.price, **kwargs)

    @no_type_check
    def vega(self, **kwargs) -> Tensor:
        """Returns delta of the derivative.

        Returns:
            torch.Tensor
        """
        return autogreek.vega(self.price, **kwargs)

    @no_type_check
    def theta(self, **kwargs) -> Tensor:
        """Returns delta of the derivative.

        Returns:
            torch.Tensor
        """
        return autogreek.theta(self.price, **kwargs)

    def inputs(self) -> List[str]:
        """Returns the names of input features.

        Returns:
            list
        """
        return list(signature(self.delta).parameters.keys())


def acquire_params_from_derivative_0(
    derivative: Optional[
        Union[
            EuropeanOption, EuropeanBinaryOption, AmericanBinaryOption, LookbackOption
        ]
    ],
    log_moneyness: Optional[Tensor] = None,
    time_to_maturity: Optional[Tensor] = None,
) -> Tuple[Tensor, Tensor]:
    if log_moneyness is None:
        if derivative is None:
            raise ValueError(
                "log_moneyness is required if derivative is not set at this initialization."
            )
        log_moneyness = derivative.log_moneyness()
    if time_to_maturity is None:
        if derivative is None:
            raise ValueError(
                "time_to_maturity is required if derivative is not set at this initialization."
            )
        time_to_maturity = derivative.time_to_maturity()
    return log_moneyness, time_to_maturity


def acquire_params_from_derivative_1(
    derivative: Optional[
        Union[
            EuropeanOption, EuropeanBinaryOption, AmericanBinaryOption, LookbackOption
        ]
    ],
    log_moneyness: Optional[Tensor] = None,
    time_to_maturity: Optional[Tensor] = None,
    volatility: Optional[Tensor] = None,
) -> Tuple[Tensor, Tensor, Tensor]:
    log_moneyness, time_to_maturity = acquire_params_from_derivative_0(
        derivative=derivative,
        log_moneyness=log_moneyness,
        time_to_maturity=time_to_maturity,
    )
    if volatility is None:
        if derivative is None:
            raise ValueError(
                "time_to_maturity is required if derivative is not set at this initialization."
            )
        if derivative.ul().volatility is None:
            raise AttributeError(
                "please simulate first and check if volatility exists in the derivative's underlier."
            )
        volatility = derivative.ul().volatility
    return log_moneyness, time_to_maturity, volatility


def acquire_params_from_derivative_2(
    derivative: Optional[Union[AmericanBinaryOption, LookbackOption]],
    log_moneyness: Optional[Tensor] = None,
    max_log_moneyness: Optional[Tensor] = None,
    time_to_maturity: Optional[Tensor] = None,
    volatility: Optional[Tensor] = None,
) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
    log_moneyness, time_to_maturity, volatility = acquire_params_from_derivative_1(
        derivative=derivative,
        log_moneyness=log_moneyness,
        time_to_maturity=time_to_maturity,
        volatility=volatility,
    )
    if max_log_moneyness is None:
        if derivative is None:
            raise ValueError(
                "max_log_moneyness is required if derivative is not set at this initialization."
            )
        max_log_moneyness = derivative.max_log_moneyness()

    return log_moneyness, max_log_moneyness, time_to_maturity, volatility



import copy
from typing import List
from typing import Optional
from typing import TypeVar
from typing import Union

import torch
from torch import Tensor
from torch.nn import Module

from pfhedge.instruments import BaseDerivative

from ._base import Feature
from ._getter import get_feature

T = TypeVar("T", bound="FeatureList")
TM = TypeVar("TM", bound="ModuleOutput")


class FeatureList(Feature):
    """Holds features in a list.

    Args:
        features (list[str | Features]): A list of features.

    Examples:

        >>> from pfhedge.features import FeatureList
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        >>>
        >>> derivative = EuropeanOption(BrownianStock(), maturity=5/250)
        >>> derivative.simulate(n_paths=2)
        >>> f = FeatureList(["moneyness", "volatility", "empty"]).of(derivative)
        >>> len(f)
        3
        >>> f.get(0).size()
        torch.Size([2, 1, 3])
    """

    def __init__(self, features: List[Union[str, Feature]]):
        self.features = list(map(get_feature, features))

    def __len__(self) -> int:
        return len(self.features)

    def get(self, time_step: Optional[int]) -> Tensor:
        # Return size: (N, T, F)
        return torch.cat([f.get(time_step) for f in self.features], dim=-1)

    def __str__(self) -> str:
        return str(list(map(str, self.features)))

    def __repr__(self) -> str:
        return str(self)

    def of(self: T, derivative: BaseDerivative, hedger: Optional[Module] = None) -> T:
        output = copy.copy(self)
        output.features = [f.of(derivative, hedger) for f in self.features]
        return output

    def is_state_dependent(self) -> bool:
        return any(map(lambda f: f.is_state_dependent(), self.features))


class ModuleOutput(Feature, Module):
    r"""The feature computed as an output of a :class:`torch.nn.Module`.

    Args:
        module (torch.nn.Module): Module to compute the value of the feature.
            The input and output shapes should be
            :math:`(N, *, H_{\mathrm{in}}) \to (N, *, H_{\mathrm{out}})` where
            :math:`N` is the number of simulated paths of the underlying instrument,
            :math:`H_{\mathrm{in}}` is the number of input features,
            :math:`H_{\mathrm{out}}` is the number of output features, and
            :math:`*` means any number of additional dimensions.
        inputs (list[Feature]): The input features to the module.

    Examples:
        >>> from torch.nn import Linear
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        >>> derivative = EuropeanOption(BrownianStock())
        >>> derivative.simulate(n_paths=3)
        >>>
        >>> m = Linear(2, 1)
        >>> f = ModuleOutput(m, inputs=["moneyness", "expiry_time"]).of(derivative)
        >>> f.get(0).size()
        torch.Size([3, 1, 1])
        >>> f
        ModuleOutput(
          inputs=['moneyness', 'expiry_time']
          (module): Linear(in_features=2, out_features=1, bias=True)
        )

        >>> from pfhedge.nn import BlackScholes
        >>>
        >>> _ = torch.manual_seed(42)
        >>> derivative = EuropeanOption(BrownianStock())
        >>> derivative.simulate(n_paths=3)
        >>> m = BlackScholes(derivative)
        >>> f = ModuleOutput(m, ["log_moneyness", "expiry_time", "volatility"])
        >>> f = f.of(derivative)
        >>> f.get(0).size()
        torch.Size([3, 1, 1])
    """

    module: Module
    inputs: FeatureList

    def __init__(self, module: Module, inputs: List[Union[str, Feature]]) -> None:
        super(Module, self).__init__()
        super(Feature, self).__init__()

        self.add_module("module", module)
        self.inputs = FeatureList(inputs)

    def extra_repr(self) -> str:
        return "inputs=" + str(self.inputs)

    def forward(self, input: Tensor) -> Tensor:
        return self.module(input)

    def get(self, time_step: Optional[int]) -> Tensor:
        return self(self.inputs.get(time_step))

    def of(self: TM, derivative: BaseDerivative, hedger: Optional[Module] = None) -> TM:
        self.inputs = self.inputs.of(derivative, hedger)
        return self

    def is_state_dependent(self) -> bool:
        return self.inputs.is_state_dependent()



from typing import Optional
from typing import Tuple

import torch
from torch import Tensor

from pfhedge._utils.typing import TensorOrScalar

from ._utils import cast_state


def generate_vasicek(
    n_paths: int,
    n_steps: int,
    init_state: Optional[Tuple[TensorOrScalar, ...]] = None,
    kappa: TensorOrScalar = 1.0,
    theta: TensorOrScalar = 0.04,
    sigma: TensorOrScalar = 0.04,
    dt: TensorOrScalar = 1 / 250,
    dtype: Optional[torch.dtype] = None,
    device: Optional[torch.device] = None,
) -> Tensor:
    r"""Returns time series following Vasicek model.

    The time evolution of the process is given by:

    .. math::

        dX(t) = \kappa (\theta - X(t)) dt + \sigma dW(t) .

    References:
        - Gillespie, D.T., 1996.
          Exact numerical simulation of the Ornstein-Uhlenbeck process and its integral.
          Physical review E, 54(2), p.2084.

    Args:
        n_paths (int): The number of simulated paths.
        n_steps (int): The number of time steps.
        init_state (tuple[torch.Tensor | float], optional): The initial state of
            the time series.
            This is specified by a tuple :math:`(X(0),)`.
            It also accepts a :class:`torch.Tensor` or a :class:`float`.
            If ``None`` (default), it uses :math:`(\theta, )`.
        kappa (torch.Tensor or float, default=1.0): The parameter :math:`\kappa`.
        theta (torch.Tensor or float, default=0.04): The parameter :math:`\theta`.
        sigma (torch.Tensor or float, default=0.04): The parameter :math:`\sigma`.
        dt (torch.Tensor or float, default=1/250): The intervals of the time steps.
        dtype (torch.dtype, optional): The desired data type of returned tensor.
            Default: If ``None``, uses a global default
            (see :func:`torch.set_default_tensor_type()`).
        device (torch.device, optional): The desired device of returned tensor.
            Default: if None, uses the current device for the default tensor type
            (see :func:`torch.set_default_tensor_type()`).
            ``device`` will be the CPU for CPU tensor types and the current CUDA device
            for CUDA tensor types.

    Shape:
        - Output: :math:`(N, T)` where
          :math:`N` is the number of paths and
          :math:`T` is the number of time steps.

    Returns:
        torch.Tensor

    Examples:
        >>> from pfhedge.stochastic import generate_vasicek
        ...
        >>> _ = torch.manual_seed(42)
        >>> generate_vasicek(2, 5)
        tensor([[0.0400, 0.0409, 0.0412, 0.0418, 0.0423],
                [0.0400, 0.0395, 0.0451, 0.0435, 0.0446]])
    """
    if init_state is None:
        init_state = (theta,)

    init_state = cast_state(init_state, dtype, device)

    if init_state[0] != 0:
        new_init_state = (init_state[0] - theta,)
        return theta + generate_vasicek(
            n_paths=n_paths,
            n_steps=n_steps,
            init_state=new_init_state,
            kappa=kappa,
            theta=0.0,
            sigma=sigma,
            dt=dt,
            dtype=dtype,
            device=device,
        )

    output = torch.empty(*(n_paths, n_steps), dtype=dtype, device=device)  # type: ignore
    output[:, 0] = init_state[0]

    # Cast to Tensor with desired dtype and device
    kappa, theta, sigma, dt = map(torch.as_tensor, (kappa, theta, sigma, dt))
    kappa, theta, sigma, dt = map(lambda t: t.to(output), (kappa, theta, sigma, dt))

    randn = torch.randn_like(output)

    # Compute \mu: Equation (3.3)
    mu = (-kappa * dt).exp()
    for i_step in range(n_steps - 1):
        # Compute \sigma_X: Equation (3.4)
        vola = sigma * ((1 - mu.square()) / 2 / kappa).sqrt()
        output[:, i_step + 1] = mu * output[:, i_step] + vola * randn[:, i_step]

    return output



from typing import Optional

from torch import Tensor
from torch.nn import Module

from pfhedge._utils.str import _format_float

from ..functional import clamp
from ..functional import leaky_clamp


class LeakyClamp(Module):
    r"""Leakily clamp all elements in ``input`` into the range :math:`[\min, \max]`.

    The bounds :math:`\min` and :math:`\max` can be tensors.

    If :math:`\min \leq \max`:

    .. math::
        \text{output} = \begin{cases}
            \min + \text{clampled_slope} * (\text{input} - \min) &
            \text{input} < \min \\
            \text{input} & \min \leq \text{input} \leq \max \\
            \max + \text{clampled_slope} * (\text{input} - \max) &
            \max < \text{input}
        \end{cases}

    If :math:`\min > \max`:

    .. math::

        \text{output} = \begin{cases}
            \frac12 (\min + \max)
            & \text{inverted_output} = \text{'mean'} \\
            \max
            & \text{inverted_output} = \text{'max'} \\
        \end{cases}

    .. seealso::
        - :func:`pfhedge.nn.functional.leaky_clamp`

    Args:
        clamped_slope (float, default=0.01):
            Controls the slope in the clampled regions.
        inverted_output ({'mean', ''max'}, default='mean'):
            Controls the output when :math:`\min > \max`.
            'max' is consistent with :func:`torch.clamp`.

    Shape:
        - input: :math:`(N, *)` where
          :math:`*` means any number of additional dimensions.
        - min: :math:`(N, *)` or any size broadcastable to ``input``.
        - max: :math:`(N, *)` or any size broadcastable to ``input``.
        - output: :math:`(N, *)`, same shape as the input.

    Examples:
        >>> import torch
        >>> from pfhedge.nn import LeakyClamp
        >>> m = LeakyClamp()
        >>> input = torch.linspace(-2, 12, 15) * 0.1
        >>> input
        tensor([-0.2000, -0.1000,  0.0000,  0.1000,  0.2000,  0.3000,  0.4000,  0.5000,
                 0.6000,  0.7000,  0.8000,  0.9000,  1.0000,  1.1000,  1.2000])
        >>> m(input, 0.0, 1.0)
        tensor([-2.0000e-03, -1.0000e-03,  0.0000e+00,  1.0000e-01,  2.0000e-01,
                 3.0000e-01,  4.0000e-01,  5.0000e-01,  6.0000e-01,  7.0000e-01,
                 8.0000e-01,  9.0000e-01,  1.0000e+00,  1.0010e+00,  1.0020e+00])
    """

    def __init__(self, clamped_slope: float = 0.01, inverted_output: str = "mean"):
        super().__init__()
        self.clamped_slope = clamped_slope
        self.inverted_output = inverted_output

    def extra_repr(self) -> str:
        return "clamped_slope=" + _format_float(self.clamped_slope)

    def forward(
        self, input: Tensor, min: Optional[Tensor] = None, max: Optional[Tensor] = None
    ) -> Tensor:
        r"""Clamp all elements in ``input`` into the range :math:`[\min, \max]`.

        Args:
            input (torch.Tensor): The input tensor.
            min (torch.Tensor, optional): Lower-bound of the range to be clamped to.
            max (torch.Tensor, optional): Upper-bound of the range to be clamped to.

        Returns:
            torch.Tensor
        """
        return leaky_clamp(input, min=min, max=max, clamped_slope=self.clamped_slope)


class Clamp(Module):
    r"""Clamp all elements in ``input`` into the range :math:`[\min, \max]`.

    The bounds :math:`\min` and :math:`\max` can be tensors.

    If :math:`\min \leq \max`:

    .. math::

        \text{output} = \begin{cases}
        \min & \text{input} < \min \\
        \text{input} & \min \leq \text{input} \leq \max \\
        \max & \max < \text{input}
        \end{cases}

    If :math:`\min > \max`:

    .. math::

        \text{output} = \begin{cases}
            \frac12 (\min + \max)
            & \text{inverted_output} = \text{'mean'} \\
            \max
            & \text{inverted_output} = \text{'max'} \\
        \end{cases}

    .. seealso::
        - :func:`torch.clamp`
        - :func:`pfhedge.nn.functional.clamp`

    Args:
        inverted_output ({'mean', ''max'}, default='mean'):
            Controls the output when :math:`\min > \max`.
            'max' is consistent with :func:`torch.clamp`.

    Shape:
        - input: :math:`(N, *)` where
          :math:`*` means any number of additional dimensions.
        - min: :math:`(N, *)` or any size broadcastable to ``input``.
        - max: :math:`(N, *)` or any size broadcastable to ``input``.
        - output: :math:`(N, *)`, same shape as the input.

    Examples:

        >>> import torch
        >>> from pfhedge.nn import Clamp
        >>>
        >>> m = Clamp()
        >>> input = torch.linspace(-2, 12, 15) * 0.1
        >>> input
        tensor([-0.2000, -0.1000,  0.0000,  0.1000,  0.2000,  0.3000,  0.4000,  0.5000,
                 0.6000,  0.7000,  0.8000,  0.9000,  1.0000,  1.1000,  1.2000])
        >>> m(input, 0.0, 1.0)
        tensor([0.0000, 0.0000, 0.0000, 0.1000, 0.2000, 0.3000, 0.4000, 0.5000, 0.6000,
                0.7000, 0.8000, 0.9000, 1.0000, 1.0000, 1.0000])

        When :math:`\min > \max`, returns the mean of :math:`\min` and :math:`\max`.

        >>> input = torch.tensor([1.0, 0.0])
        >>> min = torch.tensor([0.0, 1.0])
        >>> max = torch.tensor([0.0, 0.0])
        >>> m(input, min, max)
        tensor([0.0000, 0.5000])
    """

    def forward(
        self, input: Tensor, min: Optional[Tensor] = None, max: Optional[Tensor] = None
    ) -> Tensor:
        r"""Clamp all elements in ``input`` into the range :math:`[\min, \max]`.

        Args:
            input (torch.Tensor): The input tensor.
            min (torch.Tensor, optional): Lower-bound of the range to be clamped to.
            max (torch.Tensor, optional): Upper-bound of the range to be clamped to.

        Returns:
            torch.Tensor
        """
        return clamp(input, min=min, max=max)



from abc import ABC
from typing import Callable

import torch
from torch import Tensor
from torch.nn import Module
from torch.nn.parameter import Parameter

from pfhedge._utils.bisect import bisect
from pfhedge._utils.str import _format_float
from pfhedge._utils.typing import TensorOrScalar

from ..functional import entropic_risk_measure
from ..functional import exp_utility
from ..functional import expected_shortfall
from ..functional import isoelastic_utility
from ..functional import quadratic_cvar


class HedgeLoss(Module, ABC):
    """Base class for hedging criteria."""

    def forward(self, input: Tensor, target: TensorOrScalar = 0.0) -> Tensor:
        """Returns the loss of the profit-loss distribution.

        This method should be overridden.

        Args:
            input (torch.Tensor): The distribution of the profit and loss.
            target (torch.Tensor or float, default=0): The target portfolio to replicate.
                Typically, target is the payoff of a derivative.

        Shape:
            - input: :math:`(N, *)` where
              :math:`*` means any number of additional dimensions.
            - target: :math:`(N, *)`
            - output: :math:`(*)`

        Returns:
            torch.Tensor
        """

    def cash(self, input: Tensor, target: TensorOrScalar = 0.0) -> Tensor:
        """Returns the cash amount which is as preferable as
        the given profit-loss distribution in terms of the loss.

        The output ``cash`` is expected to satisfy the following relation:

        .. code::

            loss(torch.full_like(pl, cash)) = loss(pl)

        By default, the output is computed by binary search.
        If analytic form is known, it is recommended to override this method
        for faster computation.

        Args:
            input (torch.Tensor): The distribution of the profit and loss.
            target (torch.Tensor or float, default=0): The target portfolio to replicate.
                Typically, target is the payoff of a derivative.

        Shape:
            - input: :math:`(N, *)` where
              :math:`*` means any number of additional dimensions.
            - target: :math:`(N, *)`
            - output: :math:`(*)`

        Returns:
            torch.Tensor
        """
        pl = input - target
        return bisect(self, self(pl), pl.min(), pl.max())


class EntropicRiskMeasure(HedgeLoss):
    r"""Creates a criterion that measures
    the entropic risk measure.

    The entropic risk measure of the profit-loss distribution
    :math:`\text{pl}` is given by:

    .. math::
        \text{loss}(\text{PL}) = \frac{1}{a}
        \log(- \mathbf{E}[u(\text{PL})]) \,,
        \quad
        u(x) = -\exp(-a x) \,.

    .. seealso::
        - :func:`pfhedge.nn.functional.exp_utility`:
          The corresponding utility function.

    Args:
        a (float, default=1.0): Risk aversion coefficient of
            the exponential utility.
            This parameter should be positive.

    Shape:
        - input: :math:`(N, *)` where
          :math:`*` means any number of additional dimensions.
        - target: :math:`(N, *)`
        - output: :math:`(*)`

    Examples:
        >>> from pfhedge.nn import EntropicRiskMeasure
        ...
        >>> loss = EntropicRiskMeasure()
        >>> input = -torch.arange(4.0)
        >>> loss(input)
        tensor(2.0539)
        >>> loss.cash(input)
        tensor(-2.0539)
    """

    def __init__(self, a: float = 1.0) -> None:
        if not a > 0:
            raise ValueError("Risk aversion coefficient should be positive.")

        super().__init__()
        self.a = a

    def extra_repr(self) -> str:
        return "a=" + _format_float(self.a) if self.a != 1 else ""

    def forward(self, input: Tensor, target: TensorOrScalar = 0.0) -> Tensor:
        return entropic_risk_measure(input - target, a=self.a)

    def cash(self, input: Tensor, target: TensorOrScalar = 0.0) -> Tensor:
        return -self(input - target)


class EntropicLoss(HedgeLoss):
    r"""Creates a criterion that measures the expected exponential utility.

    The loss of the profit-loss :math:`\text{PL}` is given by:

    .. math::
        \text{loss}(\text{PL}) = -\mathbf{E}[u(\text{PL})] \,,
        \quad
        u(x) = -\exp(-a x) \,.

    .. seealso::
        - :func:`pfhedge.nn.functional.exp_utility`:
          The corresponding utility function.

    Args:
        a (float > 0, default=1.0): Risk aversion coefficient of
            the exponential utility.

    Shape:
        - input: :math:`(N, *)` where
            :math:`*` means any number of additional dimensions.
        - target: :math:`(N, *)`
        - output: :math:`(*)`

    Examples:
        >>> from pfhedge.nn import EntropicLoss
        ...
        >>> loss = EntropicLoss()
        >>> input = -torch.arange(4.0)
        >>> loss(input)
        tensor(7.7982)
        >>> loss.cash(input)
        tensor(-2.0539)
    """

    def __init__(self, a: float = 1.0) -> None:
        if not a > 0:
            raise ValueError("Risk aversion coefficient should be positive.")

        super().__init__()
        self.a = a

    def extra_repr(self) -> str:
        return "a=" + _format_float(self.a) if self.a != 1 else ""

    def forward(self, input: Tensor, target: TensorOrScalar = 0.0) -> Tensor:
        return -exp_utility(input - target, a=self.a).mean(0)

    def cash(self, input: Tensor, target: TensorOrScalar = 0.0) -> Tensor:
        return -(-exp_utility(input - target, a=self.a).mean(0)).log() / self.a


class IsoelasticLoss(HedgeLoss):
    r"""Creates a criterion that measures the expected isoelastic utility.

    The loss of the profit-loss :math:`\text{PL}` is given by:

    .. math::
        \text{loss}(\text{PL}) = -\mathbf{E}[u(\text{PL})] \,,
        \quad
        u(x) = \begin{cases}
        x^{1 - a} & a \neq 1 \\
        \log{x} & a = 1
        \end{cases} \,.

    .. seealso::
        - :func:`pfhedge.nn.functional.isoelastic_utility`:
          The corresponding utility function.

    Args:
        a (float): Relative risk aversion coefficient of the isoelastic utility.
            This parameter should satisfy :math:`0 < a \leq 1`.

    Shape:
        - input: :math:`(N, *)` where
            :math:`*` means any number of additional dimensions.
        - target: :math:`(N, *)`
        - output: :math:`(*)`

    Examples:
        >>> from pfhedge.nn import IsoelasticLoss
        ...
        >>> loss = IsoelasticLoss(0.5)
        >>> input = torch.arange(1.0, 5.0)
        >>> loss(input)
        tensor(-1.5366)
        >>> loss.cash(input)
        tensor(2.3610)

        >>> loss = IsoelasticLoss(1.0)
        >>> pl = torch.arange(1.0, 5.0)
        >>> loss(input)
        tensor(-0.7945)
        >>> loss.cash(input)
        tensor(2.2134)
    """

    def __init__(self, a: float) -> None:
        if not 0 < a <= 1:
            raise ValueError(
                "Relative risk aversion coefficient should satisfy 0 < a <= 1."
            )

        super().__init__()
        self.a = a

    def extra_repr(self) -> str:
        return "a=" + _format_float(self.a) if self.a != 1 else ""

    def forward(self, input: Tensor, target: TensorOrScalar = 0.0) -> Tensor:
        return -isoelastic_utility(input - target, a=self.a).mean(0)


class ExpectedShortfall(HedgeLoss):
    r"""Creates a criterion that measures the expected shortfall.

    .. seealso::
        - :func:`pfhedge.nn.functional.expected_shortfall`

    Args:
        p (float, default=0.1): Quantile level.
            This parameter should satisfy :math:`0 < p \leq 1`.

    Shape:
        - input: :math:`(N, *)` where
            :math:`*` means any number of additional dimensions.
        - target: :math:`(N, *)`
        - output: :math:`(*)`

    Examples:
        >>> from pfhedge.nn import ExpectedShortfall
        ...
        >>> loss = ExpectedShortfall(0.5)
        >>> input = -torch.arange(4.0)
        >>> loss(input)
        tensor(2.5000)
        >>> loss.cash(input)
        tensor(-2.5000)
    """

    def __init__(self, p: float = 0.1):
        if not 0 < p <= 1:
            raise ValueError("The quantile level should satisfy 0 < p <= 1.")

        super().__init__()
        self.p = p

    def extra_repr(self) -> str:
        return str(self.p)

    def forward(self, input: Tensor, target: TensorOrScalar = 0.0) -> Tensor:
        return expected_shortfall(input - target, p=self.p, dim=0)

    def cash(self, input: Tensor, target: TensorOrScalar = 0.0) -> Tensor:
        return -self(input - target)


class QuadraticCVaR(HedgeLoss):
    """Creates a criterion that measures the QuadraticCVaR.

    .. math::

        \\rho (X) = \\inf_\\omega \\left\\{\\omega + \\lambda || \\min\\{0, X + \\omega\\}||_2\\right\\}.

    for :math:`\lambda\geq1`.

    References:
        - Buehler, Hans, Statistical Hedging (March 1, 2019). Available at SSRN: http://dx.doi.org/10.2139/ssrn.2913250
          (See Conclusion.)

    .. seealso::
        - :func:`pfhedge.nn.functional.quadratic_cvar`

    Args:
        lam (float, default=10.0): :math:`\lambda`.
            This parameter should satisfy :math:`\lambda \geq 1`.

    Shape:
        - input: :math:`(N, *)` where
            :math:`*` means any number of additional dimensions.
        - target: :math:`(N, *)`
        - output: :math:`(*)`

    Examples:
        >>> from pfhedge.nn import QuadraticCVaR
        ...
        >>> loss = QuadraticCVaR(2.0)
        >>> input = -torch.arange(10.0)
        >>> loss(input)
        tensor(7.9750)
        >>> loss.cash(input)
        tensor(-7.9750)
    """  # NOQA

    def __init__(self, lam: float = 10.0):
        if not lam >= 1.0:
            raise ValueError("The lam should satisfy lam >= 1.")

        super().__init__()
        self.lam = lam

    def extra_repr(self) -> str:
        return str(self.lam)

    def forward(self, input: Tensor, target: TensorOrScalar = 0.0) -> Tensor:
        return quadratic_cvar(input - target, lam=self.lam, dim=0)

    def cash(self, input: Tensor, target: TensorOrScalar = 0.0) -> Tensor:
        return -self(input - target)


class OCE(HedgeLoss):
    r"""Creates a criterion that measures the optimized certainty equivalent.

    The certainty equivalent is given by:

    .. math::
        \text{loss}(X, w) = w - \mathrm{E}[u(X + w)]

    Minimization of loss gives the optimized certainty equivalent.

    .. math::
        \rho_u(X) = \inf_w \text{loss}(X, w)

    Args:
        utility (callable): Utility function.

    Attributes:
        w (torch.nn.Parameter): Represents wealth.

    Examples:
        >>> from pfhedge.nn.modules.loss import OCE
        ...
        >>> _ = torch.manual_seed(42)
        >>> m = OCE(lambda x: 1 - (-x).exp())
        >>> pl = torch.randn(10)
        >>> m(pl)
        tensor(0.0855, grad_fn=<SubBackward0>)
        >>> m.cash(pl)
        tensor(-0.0821)
    """

    def __init__(self, utility: Callable[[Tensor], Tensor]) -> None:
        super().__init__()

        self.utility = utility
        self.w = Parameter(torch.tensor(0.0))

    def extra_repr(self) -> str:
        w = float(self.w.item())
        return self.utility.__name__ + ", w=" + _format_float(w)

    def forward(self, input: Tensor, target: TensorOrScalar = 0.0) -> Tensor:
        return self.w - self.utility(input - target + self.w).mean(0)



from copy import deepcopy
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

from torch.nn import Identity
from torch.nn import LazyLinear
from torch.nn import Linear
from torch.nn import Module
from torch.nn import ReLU
from torch.nn import Sequential


class MultiLayerPerceptron(Sequential):
    r"""Creates a multilayer perceptron.

    Number of input features is lazily determined.

    Args:
        in_features (int, optional): Size of each input sample.
            If ``None`` (default), the number of input features will be
            will be inferred from the ``input.shape[-1]`` after the first call to
            ``forward`` is done. Also, before the first ``forward`` parameters in the
            module are of :class:`torch.nn.UninitializedParameter` class.
        out_features (int, default=1): Size of each output sample.
        n_layers (int, default=4): The number of hidden layers.
        n_units (int or tuple[int], default=32): The number of units in
            each hidden layer.
            If ``tuple[int]``, it specifies different number of units for each layer.
        activation (torch.nn.Module, default=torch.nn.ReLU()):
            The activation module of the hidden layers.
            Default is a :class:`torch.nn.ReLU` instance.
        out_activation (torch.nn.Module, default=torch.nn.Identity()):
            The activation module of the output layer.
            Default is a :class:`torch.nn.Identity` instance.

    Shape:
        - Input: :math:`(N, *, H_{\text{in}})` where
          :math:`*` means any number of additional dimensions and
          :math:`H_{\text{in}}` is the number of input features.
        - Output: :math:`(N, *, H_{\text{out}})` where
          all but the last dimension are the same shape as the input and
          :math:`H_{\text{out}}` is the number of output features.

    Examples:

        By default, ``in_features`` is lazily determined:

        >>> import torch
        >>> from pfhedge.nn import MultiLayerPerceptron
        >>>
        >>> m = MultiLayerPerceptron()
        >>> m
        MultiLayerPerceptron(
          (0): LazyLinear(in_features=0, out_features=32, bias=True)
          (1): ReLU()
          (2): Linear(in_features=32, out_features=32, bias=True)
          (3): ReLU()
          (4): Linear(in_features=32, out_features=32, bias=True)
          (5): ReLU()
          (6): Linear(in_features=32, out_features=32, bias=True)
          (7): ReLU()
          (8): Linear(in_features=32, out_features=1, bias=True)
          (9): Identity()
        )
        >>> _ = m(torch.zeros(3, 2))
        >>> m
        MultiLayerPerceptron(
          (0): Linear(in_features=2, out_features=32, bias=True)
          (1): ReLU()
          (2): Linear(in_features=32, out_features=32, bias=True)
          (3): ReLU()
          (4): Linear(in_features=32, out_features=32, bias=True)
          (5): ReLU()
          (6): Linear(in_features=32, out_features=32, bias=True)
          (7): ReLU()
          (8): Linear(in_features=32, out_features=1, bias=True)
          (9): Identity()
        )

        Specify different number of layers for each layer:

        >>> m = MultiLayerPerceptron(1, 1, n_layers=2, n_units=(16, 32))
        >>> m
        MultiLayerPerceptron(
          (0): Linear(in_features=1, out_features=16, bias=True)
          (1): ReLU()
          (2): Linear(in_features=16, out_features=32, bias=True)
          (3): ReLU()
          (4): Linear(in_features=32, out_features=1, bias=True)
          (5): Identity()
        )
    """

    def __init__(
        self,
        in_features: Optional[int] = None,
        out_features: int = 1,
        n_layers: int = 4,
        n_units: Union[int, Sequence[int]] = 32,
        activation: Module = ReLU(),
        out_activation: Module = Identity(),
    ):
        n_units = (n_units,) * n_layers if isinstance(n_units, int) else n_units

        layers: List[Module] = []
        for i in range(n_layers):
            if i == 0:
                if in_features is None:
                    layers.append(LazyLinear(n_units[0]))
                else:
                    layers.append(Linear(in_features, n_units[0]))
            else:
                layers.append(Linear(n_units[i - 1], n_units[i]))
            layers.append(deepcopy(activation))
        layers.append(Linear(n_units[-1], out_features))
        layers.append(deepcopy(out_activation))

        super().__init__(*layers)



from .brownian import generate_brownian  # NOQA
from .brownian import generate_geometric_brownian  # NOQA
from .cir import generate_cir  # NOQA
from .heston import generate_heston  # NOQA
from .local_volatility import generate_local_volatility_process  # NOQA
from .merton_jump import generate_merton_jump  # NOQA
from .random import randn_antithetic  # NOQA
from .random import randn_sobol_boxmuller  # NOQA
from .rough_bergomi import generate_rough_bergomi  # NOQA
from .vasicek import generate_vasicek  # NOQA



from collections import namedtuple
from typing import Optional
from typing import Tuple

import torch
from torch import Tensor

from pfhedge._utils.str import _addindent
from pfhedge._utils.typing import TensorOrScalar

from ._utils import cast_state
from .cir import generate_cir


class SpotVarianceTuple(namedtuple("SpotVarianceTuple", ["spot", "variance"])):

    __module__ = "pfhedge.stochastic"

    def __repr__(self) -> str:
        items_str_list = []
        for field, tensor in self._asdict().items():
            items_str_list.append(field + "=\n" + str(tensor))
        items_str = _addindent("\n".join(items_str_list), 2)
        return self.__class__.__name__ + "(\n" + items_str + "\n)"

    @property
    def volatility(self) -> Tensor:
        return self.variance.clamp(min=0.0).sqrt()


def generate_heston(
    n_paths: int,
    n_steps: int,
    init_state: Optional[Tuple[TensorOrScalar, ...]] = None,
    kappa: float = 1.0,
    theta: float = 0.04,
    sigma: float = 0.2,
    rho: float = -0.7,
    dt: float = 1 / 250,
    dtype: Optional[torch.dtype] = None,
    device: Optional[torch.device] = None,
) -> SpotVarianceTuple:
    """Returns time series following Heston model.

    The time evolution of the process is given by:

    .. math::

        dS(t) = S(t) \\sqrt{V(t)} dW_1(t) \\,, \\\\
        dV(t) = \\kappa (\\theta - V(t)) dt + \\sigma \\sqrt{V(t)} dW_2(t) \\,.

    The correlation between :math:`dW_1` and :math:`dW_2` is :math:`\\rho`.

    Time-series is generated by Andersen's QE-M method (See Reference for details).

    References:
        - Heston, S.L., 1993. A closed-form solution for options with stochastic volatility
          with applications to bond and currency options.
          The review of financial studies, 6(2), pp.327-343.
        - Andersen, Leif B.G., Efficient Simulation of the Heston Stochastic
          Volatility Model (January 23, 2007). Available at SSRN:
          https://ssrn.com/abstract=946405 or http://dx.doi.org/10.2139/ssrn.946404

    Args:
        n_paths (int): The number of simulated paths.
        n_steps (int): The number of time steps.
        init_state (tuple[torch.Tensor | float], optional): The initial state of
            the time series.
            This is specified by a tuple :math:`(S(0), V(0))`.
            If ``None`` (default), it uses :math:`(1.0, \\theta)`.
        kappa (float, default=1.0): The parameter :math:`\\kappa`.
        theta (float, default=0.04): The parameter :math:`\\theta`.
        sigma (float, default=2.0): The parameter :math:`\\sigma`.
        rho (float, default=-0.7): The parameter :math:`\\rho`.
        dt (float, default=1/250): The intervals of the time steps.
        dtype (torch.dtype, optional): The desired data type of returned tensor.
            Default: If ``None``, uses a global default
            (see :func:`torch.set_default_tensor_type()`).
        device (torch.device, optional): The desired device of returned tensor.
            Default: If ``None``, uses the current device for the default tensor type
            (see :func:`torch.set_default_tensor_type()`).
            ``device`` will be the CPU for CPU tensor types and the current CUDA device
            for CUDA tensor types.

    Shape:
        - spot: :math:`(N, T)` where
          :math:`N` is the number of paths and
          :math:`T` is the number of time steps.
        - variance: :math:`(N, T)`.

    Returns:
        (torch.Tensor, torch.Tensor): A namedtuple ``(spot, variance)``.

    Examples:
        >>> from pfhedge.stochastic import generate_heston
        ...
        >>> _ = torch.manual_seed(42)
        >>> spot, variance = generate_heston(2, 5)
        >>> spot
        tensor([[1.0000, 0.9941, 0.9905, 0.9846, 0.9706],
                [1.0000, 1.0031, 0.9800, 0.9785, 0.9735]])
        >>> variance
        tensor([[0.0400, 0.0408, 0.0411, 0.0417, 0.0422],
                [0.0400, 0.0395, 0.0452, 0.0434, 0.0446]])
    """
    if init_state is None:
        init_state = (1.0, theta)

    init_state = cast_state(init_state, dtype=dtype, device=device)

    GAMMA1 = 0.5
    GAMMA2 = 0.5

    variance = generate_cir(
        n_paths=n_paths,
        n_steps=n_steps,
        init_state=init_state[1:],
        kappa=kappa,
        theta=theta,
        sigma=sigma,
        dt=dt,
        dtype=dtype,
        device=device,
    )

    log_spot = torch.empty_like(variance)
    log_spot[:, 0] = init_state[0].log()
    randn = torch.randn_like(variance)

    for i_step in range(n_steps - 1):
        # Compute log S(t + 1): Eq(33)
        k0 = -rho * kappa * theta * dt / sigma
        k1 = GAMMA1 * dt * (kappa * rho / sigma - 0.5) - rho / sigma
        k2 = GAMMA2 * dt * (kappa * rho / sigma - 0.5) + rho / sigma
        k3 = GAMMA1 * dt * (1 - rho ** 2)
        k4 = GAMMA2 * dt * (1 - rho ** 2)
        v0 = variance[:, i_step]
        v1 = variance[:, i_step + 1]
        log_spot[:, i_step + 1] = (
            log_spot[:, i_step]
            + k0
            + k1 * v0
            + k2 * v1
            + (k3 * v0 + k4 * v1).sqrt() * randn[:, i_step]
        )

    return SpotVarianceTuple(log_spot.exp(), variance)



from collections import OrderedDict
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List
from typing import Tuple
from typing import Type

from torch import Tensor
from torch.nn import Module

from pfhedge.instruments import Derivative


class BlackScholesModuleFactory:

    _modules: Dict[str, Type[Module]]

    # singleton
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls)
            cls._instance._modules = OrderedDict()
        return cls._instance

    def register_module(self, name: str, cls: Type[Module]) -> None:
        self._modules[name] = cls

    def named_modules(self) -> Iterator[Tuple[str, Type[Module]]]:
        for name, module in self._modules.items():
            if module is not None:
                yield name, module

    def names(self) -> Iterator[str]:
        for name, _ in self.named_modules():
            yield name

    def features(self) -> Iterator[Type[Module]]:
        for _, module in self.named_modules():
            yield module

    def get_class(self, name: str) -> Type[Module]:
        return self._modules[name]

    def get_class_from_derivative(self, derivative: Derivative) -> Type[Module]:
        return self.get_class(derivative.__class__.__name__).from_derivative(derivative)  # type: ignore


class BlackScholes(Module):
    """Creates Black-Scholes formula module from a derivative.

    The ``forward`` method returns the Black-Scholes delta.

    Args:
        derivative (:class:`BaseDerivative`): The derivative to get
            the Black-Scholes formula.

    Shape:
        - input : :math:`(N, *, H_{\\mathrm{in}})` where
          :math:`*` means any number of additional dimensions and
          :math:`H_{\\mathrm{in}}` is the number of input features.
          See :meth:`inputs` for the names of the input features.
        - output : :math:`(N, *, 1)`,
          all but the last dimension are the same shape as the input.

    Examples:
        One can instantiate Black-Scholes module by using a derivative.
        For example, one can instantiate :class:`BSEuropeanOption` using
        a :class:`pfhedge.instruments.EuropeanOption`.
        The ``forward`` method returns delta of the derivative.

        >>> import torch
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        >>> from pfhedge.nn import BlackScholes
        >>>
        >>> derivative = EuropeanOption(BrownianStock(), strike=1.1)
        >>> m = BlackScholes(derivative)
        >>> m
        BSEuropeanOption(strike=1.1000)

        Instantiating :class:`BSLookbackOption` using a
        :class:`pfhedge.instruments.LookbackOption`.

        >>> from pfhedge.instruments import LookbackOption
        >>>
        >>> derivative = LookbackOption(BrownianStock(), strike=1.03)
        >>> m = BlackScholes(derivative)
        >>> m
        BSLookbackOption(strike=1.0300)
    """

    inputs: Callable[..., List[str]]  # inputs(self) -> List[str]
    price: Callable[..., Tensor]  # price(self, ...) -> Tensor
    delta: Callable[..., Tensor]  # delta(self, ...) -> Tensor
    gamma: Callable[..., Tensor]  # gamma(self, ...) -> Tensor
    vega: Callable[..., Tensor]  # vega(self, ...) -> Tensor
    theta: Callable[..., Tensor]  # theta(self, ...) -> Tensor

    def __new__(cls, derivative):
        return BlackScholesModuleFactory().get_class_from_derivative(derivative)



from .modules.bs.american_binary import BSAmericanBinaryOption  # NOQA
from .modules.bs.black_scholes import BlackScholes  # NOQA
from .modules.bs.european import BSEuropeanOption  # NOQA
from .modules.bs.european_binary import BSEuropeanBinaryOption  # NOQA
from .modules.bs.lookback import BSLookbackOption  # NOQA
from .modules.clamp import Clamp  # NOQA
from .modules.clamp import LeakyClamp  # NOQA
from .modules.hedger import Hedger  # NOQA
from .modules.loss import EntropicLoss  # NOQA
from .modules.loss import EntropicRiskMeasure  # NOQA
from .modules.loss import ExpectedShortfall  # NOQA
from .modules.loss import HedgeLoss  # NOQA
from .modules.loss import IsoelasticLoss  # NOQA
from .modules.loss import QuadraticCVaR  # NOQA
from .modules.mlp import MultiLayerPerceptron  # NOQA
from .modules.naked import Naked  # NOQA
from .modules.svi import SVIVariance  # NOQA
from .modules.ww import WhalleyWilmott  # NOQA



from typing import Any
from typing import Callable
from typing import Union

import torch
from torch import Tensor


def bisect(
    fn: Callable[[Tensor], Tensor],
    target: Tensor,
    lower: Union[float, Tensor],
    upper: Union[float, Tensor],
    precision: float = 1e-6,
    max_iter: int = 100000,
) -> Tensor:
    """Perform binary search over a tensor.

    The output tensor approximately satisfies the following relation:

    .. code-block::

        fn(output) = target

    Args:
        fn (callable[[Tensor], Tensor]): A monotone function.
        target (Tensor): Target of function values.
        lower (Tensor or float): Lower bound of binary search.
                                 This must be a CUDA Tensor if fn expects to input a CUDA Tensor.
        upper (Tensor or float): Upper bound of binary search.
                                 This must be a CUDA Tensor if fn expects to input a CUDA Tensor.
        precision (float, default=1e-6): Precision of output.
        max_iter (int, default 100000): If the number of iterations exceeds this
            value, abort computation and raise RuntimeError.

    Returns:
        torch.Tensor

    Raises:
        RuntimeError: If the number of iteration exceeds ``max_iter``.

    Examples:

        >>> target = torch.tensor([-1.0, 0.0, 1.0])
        >>> fn = torch.log
        >>> output = bisect(fn, target, 0.01, 10.0)
        >>> output
        tensor([0.3679, 1.0000, 2.7183])
        >>> torch.allclose(fn(output), target, atol=1e-6)
        True

        Monotone decreasing function:

        >>> fn = lambda input: -torch.log(input)
        >>> output = bisect(fn, target, 0.01, 10.0)
        >>> output
        tensor([2.7183, 1.0000, 0.3679])
        >>> torch.allclose(fn(output), target, atol=1e-6)
        True
    """
    lower, upper = map(torch.as_tensor, (lower, upper))

    if not (lower < upper).all():
        raise ValueError("condition lower < upper should be satisfied.")

    if (fn(lower) > fn(upper)).all():
        # If fn is a decreasing function
        def mf(inputs: Tensor) -> Tensor:
            return -fn(inputs)

        return bisect(mf, -target, lower, upper, precision=precision, max_iter=max_iter)

    n_iter = 0
    while torch.max(upper - lower) > precision:
        n_iter += 1
        if n_iter > max_iter:
            raise RuntimeError(f"Aborting since iteration exceeds max_iter={max_iter}.")

        m = (lower + upper) / 2
        output = fn(m)
        lower = lower.where(output >= target, m)
        upper = upper.where(output < target, m)

    return upper


def find_implied_volatility(
    pricer: Callable,
    price: Tensor,
    lower: float = 0.001,
    upper: float = 1.000,
    precision: float = 1e-6,
    max_iter: int = 100,
    **params: Any,
) -> Tensor:
    """Find implied volatility by binary search.

    Args:
        pricer (callable): Pricing formula of a derivative.
        price (Tensor): The price of the derivative.
        lower (float, default=0.001): Lower bound of binary search.
        upper (float, default=1.000): Upper bound of binary search.
        precision (float, default=1e-6): Computational precision of the implied
            volatility.
        max_iter (int, default 100): If the number of iterations exceeds this
            value, abort computation and raise RuntimeError.

    Returns:
        torch.Tensor
    """

    def fn(volatility: Tensor) -> Tensor:
        return pricer(volatility=volatility, **params)

    return bisect(
        fn,
        price,
        torch.as_tensor(lower).to(price),
        torch.as_tensor(upper).to(price),
        precision=precision,
        max_iter=max_iter,
    )



from ._getter import get_feature  # NOQA
from ._getter import list_feature_names  # NOQA
from ._getter import list_features  # NOQA
from .container import FeatureList  # NOQA
from .container import ModuleOutput  # NOQA
from .features import Barrier  # NOQA
from .features import Empty  # NOQA
from .features import ExpiryTime  # deprecated  # NOQA
from .features import LogMoneyness  # NOQA
from .features import MaxLogMoneyness  # NOQA
from .features import MaxMoneyness  # NOQA
from .features import Moneyness  # NOQA
from .features import Ones  # NOQA
from .features import PrevHedge  # NOQA
from .features import Spot  # NOQA
from .features import TimeToMaturity  # NOQA
from .features import UnderlierSpot  # NOQA
from .features import Variance  # NOQA
from .features import Volatility  # NOQA
from .features import Zeros  # NOQA



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import Iterator
from typing import Tuple
from typing import Type
from typing import Union

from ._base import Feature


class FeatureFactory:

    _features: Dict[str, Type[Feature]]

    # singleton
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls)
            cls._instance._features = OrderedDict()
        return cls._instance

    def register_feature(self, name: str, cls: Type[Feature]) -> None:
        """Adds a feature to the factory.

        Args:
            name (str): name of the feature.
            cls (type(Feature)): feature class to be registered.
        """
        self._features[name] = cls

    def named_features(self) -> Iterator[Tuple[str, Type[Feature]]]:
        """Returns an iterator over feature classes, yielding both the
        name of the feature class as well as the feature class itself.

        Yields:
            (string, type(Feature)): Tuple containing
                the name and feature class.
        """
        for name, feature in self._features.items():
            if feature is not None:
                yield name, feature

    def names(self) -> Iterator[str]:
        """Returns an iterator over the names of the feature classes.

        Yields:
            str: name of the feature class.
        """
        for name, _ in self.named_features():
            yield name

    def features(self) -> Iterator[Type[Feature]]:
        """Returns an iterator over feature classes.

        Yields:
            type(Feature): Feature class.
        """
        for _, feature in self.named_features():
            yield feature

    def get_class(self, name: str) -> Type[Feature]:
        """Returns the feature class with the given name.

        Parameters:
            name (str): name of the feature class.

        Returns:
            type(Feature): feature class.
        """
        if name not in self.names():
            raise KeyError(
                f"{name} is not a valid name. "
                "Use pfhedge.features.list_feature_names() to see available names."
            )
        return self._features[name]

    def get_instance(self, name: str, **kwargs: Any) -> Feature:
        """Returns the feature with the given name.

        Parameters:
            name (str): name of the feature class.

        Returns:
            Feature: feature.
        """
        return self.get_class(name)(**kwargs)  # type: ignore


def get_feature(feature: Union[str, Feature], **kwargs: Any) -> Feature:
    """Get feature from name.

    Args:
        name (str): Name of feature.
        **kwargs: Keyword arguments to pass to feature constructor.

    Returns:
        Feature

    Examples:
        >>> from pfhedge.features import get_feature
        ...
        >>> get_feature("moneyness")
        <pfhedge.features.features.Moneyness object at ...>
    """
    if isinstance(feature, str):
        feature = FeatureFactory().get_instance(feature, **kwargs)
    elif not isinstance(feature, Feature):
        raise TypeError(f"{feature} is not an instance of Feature.")
    return feature


def list_feature_dict() -> dict:
    return dict(FeatureFactory().named_features())


def list_feature_names() -> list:
    """Returns the list of the names of available features.

    Returns:
        list[str]
    """
    return sorted(list(FeatureFactory().names()))


def list_features() -> list:
    return list(FeatureFactory().features())



from typing import List

from torch import Tensor
from torch.nn import Module

from pfhedge._utils.str import _format_float
from pfhedge.nn.functional import ww_width

from ...instruments import Derivative
from .bs.black_scholes import BlackScholes


class WhalleyWilmott(Module):
    r"""Creates a module for Whalley-Wilmott's hedging strategy.

    The ``forward`` method returns the next hedge ratio.

    This is the no-transaction band strategy
    that is optimal under the premises of
    asymptotically small transaction cost, European option, and exponential utility.
    The half-width of the no-transaction band is given by

    .. math::
        w = \left( \frac{3 c \Gamma^2 S}{2 a} \right)^{1 / 3} \,,

    where :math:`c` is the transaction cost rate,
    :math:`\Gamma` is the gamma of the derivative,
    :math:`S` is the spot price of the underlying instrument, and
    :math:`a` is the risk-aversion coefficient of the exponential utility.

    Note:
        A backward computation for this module generates ``nan``
        if the :math:`\Gamma` of the derivative is too small.
        This is because the output is proportional to :math:`\Gamma^{2 / 3}`
        of which gradient diverges for :math:`\Gamma \to 0`.
        A ``dtype`` with higher precision may alleviate this problem.

    References:
        - Davis, M.H., Panas, V.G. and Zariphopoulou, T., 1993.
          European option pricing with transaction costs.
          SIAM Journal on Control and Optimization, 31(2), pp.470-493.
        - Whalley, A.E. and Wilmott, P., An asymptotic analysis of an optimal hedging
          model for option pricing with transaction costs. Mathematical Finance,
          1997, 7, 307–324.

    Args:
        derivative (:class:`pfhedge.instruments.BaseDerivative`): Derivative to hedge.
        a (float, default=1.0): Risk aversion parameter in exponential utility.

    Shape:
        - Input: :math:`(N, *, H_{\text{in}})` where
          :math:`*` means any number of additional dimensions and
          :math:`H_{\text{in}}` is the number of input features.
          See :meth:`inputs()` for the names of input features.
        - Output: :math:`(N, *, 1)`.

    Examples:
        An example for :class:`pfhedge.instruments.EuropeanOption`.

        >>> import torch
        >>> from pfhedge.nn import WhalleyWilmott
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        >>> derivative = EuropeanOption(BrownianStock(cost=1e-5))
        >>>
        >>> m = WhalleyWilmott(derivative)
        >>> m.inputs()
        ['log_moneyness', 'time_to_maturity', 'volatility', 'prev_hedge']
        >>> input = torch.tensor([
        ...     [-0.05, 0.1, 0.2, 0.5],
        ...     [-0.01, 0.1, 0.2, 0.5],
        ...     [ 0.00, 0.1, 0.2, 0.5],
        ...     [ 0.01, 0.1, 0.2, 0.5],
        ...     [ 0.05, 0.1, 0.2, 0.5]])
        >>> m(input)
        tensor([[0.2946],
                [0.5000],
                [0.5000],
                [0.5000],
                [0.7284]])

        An example for :class:`pfhedge.instruments.EuropeanOption` without cost.

        >>> derivative = EuropeanOption(BrownianStock())
        >>> m = WhalleyWilmott(derivative)
        >>> m.inputs()
        ['log_moneyness', 'time_to_maturity', 'volatility', 'prev_hedge']
        >>> input = torch.tensor([
        ...     [-0.05, 0.1, 0.2, 0.5],
        ...     [-0.01, 0.1, 0.2, 0.5],
        ...     [ 0.00, 0.1, 0.2, 0.5],
        ...     [ 0.01, 0.1, 0.2, 0.5],
        ...     [ 0.05, 0.1, 0.2, 0.5]])
        >>> m(input)
        tensor([[0.2239],
                [0.4497],
                [0.5126],
                [0.5752],
                [0.7945]])
    """

    def __init__(self, derivative: Derivative, a: float = 1.0) -> None:
        super().__init__()
        self.derivative = derivative
        self.a = a

        self.bs = BlackScholes(derivative)

    def inputs(self) -> List[str]:
        """Returns the names of input features.

        Returns:
            list[str]
        """
        return self.bs.inputs() + ["prev_hedge"]

    def extra_repr(self) -> str:
        return "a=" + _format_float(self.a) if self.a != 1 else ""

    def forward(self, input: Tensor) -> Tensor:
        prev_hedge = input[..., [-1]]

        delta = self.bs(input[..., :-1])
        width = self.width(input[..., :-1])
        min = delta - width
        max = delta + width

        return prev_hedge.clamp(min=min, max=max)

    def width(self, input: Tensor) -> Tensor:
        r"""Returns half-width of the no-transaction band.

        Args:
            input (torch.Tensor): The input tensor.

        Shape:
            - Input: :math:`(N, *, H_{\text{in}} - 1)` where
              :math:`*` means any number of additional dimensions and
              :math:`H_{\text{in}}` is the number of input features.
              See :meth:`inputs()` for the names of input features.
            - Output: :math:`(N, *, 1)`

        Returns:
            torch.Tensor
        """
        cost = self.derivative.underlier.cost
        spot = self.derivative.strike * input[..., [0]].exp()
        gamma = self.bs.gamma(*(input[..., [i]] for i in range(input.size(-1))))
        # width = (cost * (3 / 2) * gamma.square() * spot / self.a).pow(1 / 3)

        return ww_width(gamma=gamma, spot=spot, cost=cost, a=self.a)



import math
from typing import Optional
from typing import Tuple

import torch

from pfhedge._utils.typing import TensorOrScalar
from pfhedge.stochastic._utils import cast_state
from pfhedge.stochastic.heston import SpotVarianceTuple


def generate_rough_bergomi(
    n_paths: int,
    n_steps: int,
    init_state: Optional[Tuple[TensorOrScalar, ...]] = None,
    alpha: float = -0.4,
    rho: float = -0.9,
    eta: float = 1.9,
    xi: float = 0.04,
    dt: float = 1 / 250,
    dtype: Optional[torch.dtype] = None,
    device: Optional[torch.device] = None,
) -> SpotVarianceTuple:
    r"""Returns time series following the rough Bergomi (rBergomi) model.

    The time evolution of the process is given by:

    .. math::

        S(t) &=& \exp{\left\{\int_0^t \sqrt{V(u)} dB(u) - \frac{1}{2} \int_0^t V(u) du \right\}}, \\
        B(u) &=& \rho W_u^1 + \sqrt{1-\rho^2}W_u^2, \\
        V(t) &=& \xi\exp{\left\{\eta Y^\alpha(t) - \frac{\eta^2}{2} t^{2\alpha + 1}\right\}}, \\
        Y^\alpha(t) &=& \sqrt{1\alpha + 1}\int_0^t (t-u)^\alpha dW_u^1,

    :math:`dW^1` and :math:`dW^2` are the Brownian motion.

    Time-series is generated by Ryan et al.'s Monte Carlo algorithm.

    References:
        - Bayer, C., Friz, P., & Gatheral, J. (2015). Pricing under rough volatility.
          Quantitative Finance, 16(6), 887–904. https://doi.org/10.1080/14697688.2015.1099717
        - McCrickerd, R., & Pakkanen, M. S. (2018). Turbocharging Monte Carlo pricing for the rough Bergomi model.
          Quantitative Finance, 18(11), 1877–1886. https://doi.org/10.1080/14697688.2018.1459812
          Code: https://github.com/ryanmccrickerd/rough_bergomi

    Args:
        n_paths (int): The number of simulated paths.
        n_steps (int): The number of time steps.
        init_state (tuple[torch.Tensor | float], optional): The initial state of
            the time series.
            This is specified by a tuple :math:`(S(0), V(0))`.
            If ``None`` (default), it uses :math:`(1.0, \\xi)`.
        alpha (float, default=-0.4): The parameter :math:`\\alpha`.
        rho (float, default=-0.9): The parameter :math:`\\rho`.
        eta (float, default=1.9): The parameter :math:`\\eta`.
        xi (float, default=0.04): The parameter :math:`\\xi`.
        dt (float, default=1 / 250): The intervals of the time steps.
        dtype (torch.dtype, optional): The desired data type of returned tensor.
            Default: If ``None``, uses a global default
            (see :func:`torch.set_default_tensor_type()`).
        device (torch.device, optional): The desired device of returned tensor.
            Default: If ``None``, uses the current device for the default tensor type
            (see :func:`torch.set_default_tensor_type()`).
            ``device`` will be the CPU for CPU tensor types and the current CUDA device
            for CUDA tensor types.

    Shape:
        - spot: :math:`(N, T)` where
          :math:`N` is the number of paths and
          :math:`T` is the number of time steps.
        - variance: :math:`(N, T)`.

    Returns:
        (torch.Tensor, torch.Tensor): A namedtuple ``(spot, variance)``.

    Examples:
        >>> from pfhedge.stochastic import generate_rough_bergomi
        ...
        >>> _ = torch.manual_seed(42)
        >>> outputs = generate_rough_bergomi(2, 5)
        >>> outputs.spot
        tensor([[1.0000, 0.9807, 0.9563, 0.9540, 0.9570],
                [1.0000, 1.0147, 1.0097, 1.0107, 1.0164]])
        >>> outputs.variance
        tensor([[0.0400, 0.3130, 0.0105, 0.0164, 0.0068],
                [0.0400, 0.0396, 0.0049, 0.0064, 0.0149]])

    """

    if init_state is None:
        init_state = (1.0, xi)

    init_state = cast_state(init_state, dtype=dtype, device=device)
    alpha_tensor, rho_tensor, eta_tensor = cast_state(
        (alpha, rho, eta), dtype=dtype, device=device
    )

    _dW1_cov1 = dt ** (alpha + 1) / (alpha + 1)
    _dW1_cov2 = dt ** (2 * alpha + 1) / (2 * alpha + 1)
    _dW1_covariance_matrix = torch.as_tensor(
        [[dt, _dW1_cov1], [_dW1_cov1, _dW1_cov2]], dtype=dtype, device=device
    )
    _dW1_generator = torch.distributions.multivariate_normal.MultivariateNormal(
        loc=torch.as_tensor([0.0, 0.0], dtype=dtype, device=device),
        covariance_matrix=_dW1_covariance_matrix,
    )

    dW1 = _dW1_generator.sample(torch.Size([n_paths, n_steps - 1]))
    dW2 = torch.randn([n_paths, n_steps - 1], dtype=dtype, device=device) * math.sqrt(
        dt
    )

    _Y1 = torch.cat([torch.zeros_like(dW1[:, :1, 1]), dW1[:, :, 1]], dim=-1)

    def discrete_TBSS_fn(k: torch.Tensor, a: TensorOrScalar) -> torch.Tensor:
        return ((k ** (a + 1) - (k - 1) ** (a + 1)) / (a + 1)) ** (1 / a)

    _gamma = (
        discrete_TBSS_fn(torch.arange(2, n_steps, dtype=dtype, device=device), alpha)
        / (n_steps - 1)
    ) ** alpha
    _gamma = torch.cat([torch.zeros(2, dtype=dtype, device=device), _gamma], dim=0)
    _Xi = dW1[:, :, 0]
    _GXi_convolve = torch.nn.functional.conv1d(
        _gamma.flip(0)[None, None, :],
        _Xi[:, None, :],
        padding=_Xi.size(1) - 1,
    )[0, :, :]
    _Y2 = _GXi_convolve[:, -n_steps:].flip(1)
    Y = torch.sqrt(2 * alpha_tensor + 1) * (_Y1 + _Y2)
    dB = rho_tensor * dW1[:, :, 0] + torch.sqrt(1 - rho_tensor.square()) * dW2
    variance = init_state[1] * torch.exp(
        eta_tensor * Y
        - 0.5
        * eta_tensor.square()
        * (torch.arange(n_steps, dtype=dtype, device=device) * dt)
        ** (2 * alpha_tensor + 1)
    )

    _increments = variance[:, :-1].sqrt() * dB - 0.5 * variance[:, :-1] * dt
    _integral = torch.cumsum(_increments, dim=1)
    log_return = torch.cat([torch.zeros_like(_integral[..., :1]), _integral], dim=-1)
    prices = init_state[0] * log_return.exp()

    return SpotVarianceTuple(prices, variance)



from torch import Tensor
from torch.nn import Module

from pfhedge._utils.typing import TensorOrScalar
from pfhedge.nn.functional import svi_variance


class SVIVariance(Module):
    r"""Returns total variance in the SVI model.

    The total variance for log strike :math:`k = \log(K / S)`,
    where :math:`K` and :math:`S` are strike and spot, reads:

    .. math::
        w = a + b \left[ \rho (k - m) + \sqrt{(k - m)^2 + \sigma^2} \right] .

    References:
        - Jim Gatheral and Antoine Jacquier,
          Arbitrage-free SVI volatility surfaces.
          [arXiv:`1204.0646 <https://arxiv.org/abs/1204.0646>`_ [q-fin.PR]]

    Args:
        a (torch.Tensor or float): The parameter :math:`a`.
        b (torch.Tensor or float): The parameter :math:`b`.
        rho (torch.Tensor or float): The parameter :math:`\rho`.
        m (torch.Tensor or float): The parameter :math:`m`.
        sigma (torch.Tensor or float): The parameter :math:`\sigma`.

    Examples:
        >>> import torch
        >>>
        >>> a, b, rho, m, sigma = 0.03, 0.10, 0.10, 0.00, 0.10
        >>> module = SVIVariance(a, b, rho, m, sigma)
        >>> input = torch.tensor([-0.10, -0.01, 0.00, 0.01, 0.10])
        >>> module(input)
        tensor([0.0431, 0.0399, 0.0400, 0.0401, 0.0451])
    """

    def __init__(
        self,
        a: TensorOrScalar,
        b: TensorOrScalar,
        rho: TensorOrScalar,
        m: TensorOrScalar,
        sigma: TensorOrScalar,
    ) -> None:
        super().__init__()
        self.a = a
        self.b = b
        self.rho = rho
        self.m = m
        self.sigma = sigma

    def forward(self, input: Tensor) -> Tensor:
        return svi_variance(
            input, a=self.a, b=self.b, rho=self.rho, m=self.m, sigma=self.sigma
        )

    def extra_repr(self) -> str:
        params = (
            f"a={self.a}",
            f"b={self.b}",
            f"rho={self.rho}",
            f"m={self.m}",
            f"sigma={self.sigma}",
        )
        return ", ".join(params)



from typing import Optional

import torch
from torch import Tensor

from pfhedge._utils.doc import _set_attr_and_docstring
from pfhedge._utils.doc import _set_docstring
from pfhedge._utils.str import _format_float
from pfhedge.nn.functional import realized_variance

from ..primary.base import BasePrimary
from .base import BaseDerivative


class VarianceSwap(BaseDerivative):
    r"""Variance swap.

    The payoff of a variance swap is given by

    .. math::

        \mathrm{payoff} = \sigma^2 - K ,

    where
    :math:`\sigma^2` is the realized variance of the underlying asset until maturity and
    :math:`K` is the strike.

    See :func:`pfhedge.nn.functional.realized_variance`
    for the definition of the realized variance.

    .. seealso::
        :func:`pfhedge.nn.functional.realized_variance`

    Args:
        underlier (:class:`BasePrimary`): The underlying instrument.
        strike (float, default=0.04): The strike variance of the swap.
        maturity (float, default=20/250): The maturity of the derivative.

    Attributes:
        dtype (torch.dtype): The dtype with which the simulated time-series are
            represented.
        device (torch.device): The device where the simulated time-series are.

    Examples:
        >>> import torch
        >>> from pfhedge.nn.functional import realized_variance
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import VarianceSwap
        >>>
        >>> _ = torch.manual_seed(42)
        >>> derivative = VarianceSwap(BrownianStock(), strike=0.04, maturity=5/250)
        >>> derivative.simulate(n_paths=2)
        >>> derivative.underlier.spot
        tensor([[1.0000, 1.0016, 1.0044, 1.0073, 0.9930, 0.9906],
                [1.0000, 0.9919, 0.9976, 1.0009, 1.0076, 1.0179]])
        >>> realized_variance(derivative.ul().spot, dt=derivative.ul().dt)
        tensor([0.0114, 0.0129])
        >>> derivative.payoff()
        tensor([-0.0286, -0.0271])
    """

    def __init__(
        self,
        underlier: BasePrimary,
        strike: float = 0.04,
        maturity: float = 20 / 250,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
    ) -> None:
        super().__init__()
        self.register_underlier("underlier", underlier)
        self.strike = strike
        self.maturity = maturity

        # TODO(simaki): Remove later. Deprecated for > v0.12.3
        if dtype is not None or device is not None:
            self.to(dtype=dtype, device=device)
            raise DeprecationWarning(
                "Specifying device and dtype when constructing a Derivative is deprecated."
                "Specify them in the constructor of the underlier instead."
            )

    def extra_repr(self) -> str:
        return ", ".join(
            (
                "strike=" + _format_float(self.strike),
                "maturity=" + _format_float(self.maturity),
            )
        )

    def payoff_fn(self) -> Tensor:
        return realized_variance(self.ul().spot, dt=self.ul().dt) - self.strike


# Assign docstrings so they appear in Sphinx documentation
_set_attr_and_docstring(VarianceSwap, "simulate", BaseDerivative.simulate)
_set_attr_and_docstring(VarianceSwap, "to", BaseDerivative.to)
_set_attr_and_docstring(VarianceSwap, "ul", BaseDerivative.ul)
_set_attr_and_docstring(VarianceSwap, "list", BaseDerivative.list)
_set_docstring(VarianceSwap, "payoff", BaseDerivative.payoff)



from .base import BaseInstrument  # NOQA
from .base import Instrument  # NOQA
from .derivative.american_binary import AmericanBinaryOption  # NOQA
from .derivative.base import BaseDerivative  # NOQA
from .derivative.base import BaseOption  # NOQA
from .derivative.base import Derivative  # NOQA
from .derivative.base import OptionMixin  # NOQA
from .derivative.cliquet import EuropeanForwardStartOption  # NOQA
from .derivative.european import EuropeanOption  # NOQA
from .derivative.european_binary import EuropeanBinaryOption  # NOQA
from .derivative.lookback import LookbackOption  # NOQA
from .derivative.variance_swap import VarianceSwap  # NOQA
from .primary.base import BasePrimary  # NOQA
from .primary.base import Primary  # NOQA
from .primary.brownian import BrownianStock  # NOQA
from .primary.cir import CIRRate  # NOQA
from .primary.heston import HestonStock  # NOQA
from .primary.local_volatility import LocalVolatilityStock  # NOQA
from .primary.merton_jump import MertonJumpStock  # NOQA
from .primary.rough_bergomi import RoughBergomiStock  # NOQA
from .primary.vasicek import VasicekRate  # NOQA



import torch


def _addindent(string: str, n_spaces: int = 2) -> str:
    return "\n".join(" " * n_spaces + line for line in string.split("\n"))


def _format_float(value: float) -> str:
    """
    >>> _format_float(1)
    '1'
    >>> _format_float(1.0)
    '1.'
    >>> _format_float(1e-4)
    '1.0000e-04'
    """
    # format a float following PyTorch printoptions
    # see `torch.set_printoptions` for details
    tensor = torch.tensor([value])
    return torch._tensor_str._Formatter(tensor).format(value)



import math
from typing import Callable
from typing import Optional
from typing import Tuple
from typing import Union

import torch
from torch import Tensor

from pfhedge._utils.typing import TensorOrScalar
from pfhedge.stochastic._utils import cast_state


def generate_merton_jump(
    n_paths: int,
    n_steps: int,
    init_state: Union[Tuple[TensorOrScalar, ...], TensorOrScalar] = (1.0,),
    mu: float = 0.0,
    sigma: float = 0.2,
    jump_per_year: float = 68.2,
    jump_mean: float = 0.0,
    jump_std: float = 0.02,
    dt: float = 1 / 250,
    dtype: Optional[torch.dtype] = None,
    device: Optional[torch.device] = None,
    engine: Callable[..., Tensor] = torch.randn,
) -> Tensor:
    r"""Returns time series following the Merton's Jump Diffusion Model.

    The time evolution of the process is given by:

    .. math::

        \frac{dS(t)}{S(t)} = (\mu - \lambda k) dt + \sigma dW(t) + dJ(t) .

    Reference:
     - Merton, R.C., Option pricing when underlying stock returns are discontinuous,
       Journal of Financial Economics, 3 (1976), 125-144.
     - Gugole, N. (2016). Merton jump-diffusion model versus the black and scholes approach for the log-returns and volatility
       smile fitting. International Journal of Pure and Applied Mathematics, 109(3), 719-736.
       (Especially p.730 for parameter calibulation.)

    Args:
        n_paths (int): The number of simulated paths.
        n_steps (int): The number of time steps.
        init_state (tuple[torch.Tensor | float], default=(1.0,)): The initial state of
            the time series.
            This is specified by a tuple :math:`(S(0),)`.
            It also accepts a :class:`torch.Tensor` or a :class:`float`.
        mu (float, default=0.0): The parameter :math:`\mu`,
            which stands for the dirft coefficient of the time series.
        sigma (float, default=0.2): The parameter :math:`\sigma`,
            which stands for the volatility of the time series.
        jump_per_year (float, default=68.2): The average number of jumps per year.
        jump_mean (float, default=0.0): The mean of jump sizes.
        jump_std (float, default=0.02): The standard deviation of jump sizes.
        dt (float, default=1/250): The time step interval.
        dtype (torch.dtype, optional): The desired data type of returned tensor.
            Default: If ``None``, uses a global default
            (see :func:`torch.set_default_tensor_type()`).
        device (torch.device, optional): The desired device of returned tensor.
            Default: If ``None``, uses the current device for the default tensor type
            (see :func:`torch.set_default_tensor_type()`).
            ``device`` will be the CPU for CPU tensor types and the current CUDA device
            for CUDA tensor types.
        engine (callable, default=torch.randn): The desired generator of random numbers
            from a standard normal distribution.
            A function call ``engine(size, dtype=None, device=None)``
            should return a tensor filled with random numbers
            from a standard normal distribution.

    Shape:
        - Output: :math:`(N, T)` where
          :math:`N` is the number of paths and
          :math:`T` is the number of time steps.

    Returns:
        torch.Tensor

    Examples:
        >>> from pfhedge.stochastic import generate_merton_jump
        >>>
        >>> _ = torch.manual_seed(42)
        >>> generate_merton_jump(2, 5)
        tensor([[1.0000, 0.9904, 1.0074, 1.0160, 1.0117],
                [1.0000, 1.0034, 1.0040, 1.0490, 1.0457]])
    """
    # https://www.codearmo.com/python-tutorial/merton-jump-diffusion-model-python
    init_state = cast_state(init_state, dtype=dtype, device=device)

    poisson = torch.distributions.poisson.Poisson(rate=jump_per_year * dt)
    n_jumps = poisson.sample((n_paths, n_steps - 1)).to(dtype=dtype, device=device)

    # Eq. (3) in https://www.imes.boj.or.jp/research/papers/japanese/kk22-b1-3.pdf
    jump = (
        jump_mean * n_jumps
        + engine(*(n_paths, n_steps - 1), dtype=dtype, device=device)
        * jump_std
        * n_jumps.sqrt()
    )
    jump = torch.cat(
        [torch.zeros((n_paths, 1), dtype=dtype, device=device), jump], dim=1
    )

    randn = engine(*(n_paths, n_steps), dtype=dtype, device=device)
    randn[:, 0] = 0.0
    drift = (
        (mu - (sigma ** 2) / 2 - jump_per_year * (jump_mean + jump_std ** 2 / 2))
        * dt
        * torch.arange(n_steps).to(randn)
    )
    brown = randn.cumsum(1) * math.sqrt(dt)
    return init_state[0] * (drift + sigma * brown + jump.cumsum(1)).exp()
